//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "kernel.hpp"
#include "scene/scene.hpp"

#include "misc/log.hpp"
#include "misc/data.hpp"
#include "renderer.hpp"
#include "base.hpp"
#include "misc/factory.hpp"

#include "scene/background/background.hpp"
#include "scene/primitive.hpp"
#include "scene/camera.hpp"
#include "scene/batch.hpp"
#include "material/material.hpp"
#include "material/image.hpp"
#include "material/shader.hpp"
#include "material/texture.hpp"

//==============================================================================
//  Kernel::Kernel()
//==============================================================================
Kernel::Kernel() {
    base = NULL;
    renderer = NULL;
	m_scene = NULL;
}

//==============================================================================
//  Kernel::~Kernel()
//==============================================================================
Kernel::~Kernel() {
    free_data();
}

//==============================================================================
//  Kernel::free_data()
//==============================================================================
void Kernel::free_data() {
    if(base) {
        base->close();
    }
    if(renderer) {
        renderer->close();
    }
    if(m_scene) {
        delete m_scene;
    }
       
	Factory_primitive.free_data();
	Factory_background.free_data();		
	Factory_texture.free_data();
	Factory_material.free_data();
	Factory_image.free_data();	
	Factory_camera.free_data();
	Factory_controller.free_data();	
	Factory_base.free_data();
	Factory_renderer.free_data();
	Factory_batch.free_data();
	Factory_shader.free_data();	
}

//==============================================================================
//  Kernel::read_from_file()
//==============================================================================
bool Kernel::read_from_file( const std::string& name ) {
    log_function << "Initializate kernel from " << name << std::endl;

    free_data();    
    Data_node* data = new Data_node();

    if( !data->parse_file( name) ) {
        delete data;
        return false;
    }        

    if( Data_value* value = data->get_value( "base" ) ) {
        base = Factory_base.create( value );
    } else {
         log_fatal << "Base is not defined!" << std::endl;
         return false;
	}
	if( Data_value* value = data->get_value( "renderer" ) ) {
        renderer = Factory_renderer.create( value );
        
	    base->init_window(  base->get_window_width(), base->get_window_height(), 
                            base->get_fullscreen(), renderer->get_hardware_acc() );
                            
        renderer->set_viewport( 0, 0, base->get_window_width(), 
                                base->get_window_height());    
		renderer->init();
		renderer->clear_buffer( BUFFER_DEPTH | BUFFER_COLOR );
		base->swap_buffers();
    } else {
         log_fatal << "Renderer is not defined!" << std::endl;
         return false;
	}
    if( Data_value* value = data->get_value( "scene" ) ) {
        m_scene = Factory_scene.create( value );        
    } else {
         log_fatal << "Scene is not defined!" << std::endl;
         return false;
	}

	m_scene->first_update();

    delete data;
        
    log_success << "Kernel initialization successful" << std::endl;
    
    return true;
}

//==============================================================================
//  Kernel::render()
//==============================================================================
void Kernel::render() {
    m_scene->render();
    renderer->render_debug();
    base->swap_buffers();
	renderer->next_frame();
}

//==============================================================================
//  Kernel::update()
//==============================================================================
bool Kernel::update() {
    m_scene->update();    
    return base->update();
}

//==============================================================================
//  Kernel::start_renderer()
//==============================================================================
void Kernel::start_renderer() {
    renderer->start_renderer();
}

//==============================================================================
//  Kernel::stop_renderer()
//==============================================================================
void Kernel::stop_renderer() {
    renderer->stop_renderer();    
}
