//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "texture.hpp"
#include "image.hpp"
#include "../misc/log.hpp"
#include "../misc/factory.hpp"
   
//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_texture  = Factory_texture.reg<Texture>( "texture_2d" ); 
bool fac_cube = Factory_texture.reg<Texture>( "texture_cube" ); 

//==============================================================================
//  Texture::Texture()
//==============================================================================
Texture::Texture() {
    /* default */
    m_tex_id = 0;
    m_type = BYTE; 
    m_dimension = TEX_2D;
    m_use_nearest = false;
    m_use_clamp = false;
}

//==============================================================================
//  Texture::~Texture()
//==============================================================================
Texture::~Texture() {
    clear();
}

//==============================================================================
//  Texture::init_from_file()
//==============================================================================
void Texture::init(Data_node* data) {
    int width=64;
    int height=64;
    std::string f;
    
	const char *cube_images[] = {"file_rt","file_lf","file_fr","file_bk","file_up","file_dn"};
    
    data->get_value( "width", width );
    data->get_value( "height", height );
    data->get_value( "clamp", m_use_clamp );
    
    Image* image;
    if( data->get_value( "file", f ) ) {
		image = Factory_image.create( "image", f );
        image->load_from_file(f);
        set_image( image );
    } else if( data->get_value( cube_images[0], f ) ) {
	  	set_num_images( 6 );
	    m_dimension = TEX_CUBE;
			  
		for( int i=0; i<6; ++i ) {
			data->get_value( cube_images[i], f ); 
     	  	image = Factory_image.create( "image", f );
		    image->load_from_file_pfm(f);
		    set_image( i, image );   
		}		
    } else {
  		image = Factory_image.create( "image" );
        image->create_empty(width, height, Image::RGB, Image::CHAR);
        image->fill_as_default();
        set_image( image );
    }    
}

//==============================================================================
//  Texture::set_image()
//==============================================================================
void Texture::set_image(Image* image) {
    clear();
    m_images.push_back(image);
}

//==============================================================================
//  Texture::set_image()
//==============================================================================
void Texture::set_image(int index, Image* image) {
    if(index >= m_images.size()) {
        log_warning << "Illegal argument to Texture::set_image(): " << index << std::endl;
        return;
    }
    m_images[index] = image;
}

//==============================================================================
//  Texture::set_num_images()
//==============================================================================
void Texture::set_num_images(int count) {
    clear();
    m_images.resize(count);
    
    for(int i=0; i<count; i++) {
        m_images[i] = NULL;
    }
}

//==============================================================================
//  Texture::clear()
//==============================================================================
void Texture::clear() {
    m_images.resize(0);    
}

//==============================================================================
//  get_texture_from_file()
//==============================================================================
Texture* get_texture_from_file( const std::string& name ) {
    Texture* texture = Factory_texture.create("texture_2d", name);
    
    Image* image = get_image_from_file( name );	
    texture->set_image( image );
    
    return texture;
}
