//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "renderer.hpp"
#include "base.hpp"
#include "material/material.hpp"
#include "material/texture.hpp"
#include "material/image.hpp"
#include "scene/camera.hpp"
#include "misc/log.hpp"
#include "misc/factory.hpp"

Renderer* renderer = (Renderer* )NULL;

// Mask constants
int BUFFER_DEPTH;
int BUFFER_COLOR;

// Blending constants
int BLEND_ZERO;
int BLEND_ONE;
int BLEND_SRC_COLOR;
int BLEND_ONE_MINUS_SRC_COLOR;
int BLEND_DST_COLOR;
int BLEND_ONE_MINUS_DST_COLOR;
int BLEND_SRC_ALPHA;
int BLEND_ONE_MINUS_SRC_ALPHA;
int BLEND_DST_ALPHA;
int BLEND_ONE_MINUS_DST_ALPHA;
int BLEND_SRC_ALPHA_SATURATE;

int BLEND_ADD;
int BLEND_SUBTRACT;
int BLEND_REVERSE_SUBTRACT;
int BLEND_MIN;
int BLEND_MAX;

// Depth testing constants
int DEPTH_NEVER;
int DEPTH_LESS;
int DEPTH_EQUAL;
int DEPTH_LEQUAL;
int DEPTH_GREATER;
int DEPTH_NOTEQUAL;
int DEPTH_GEQUAL;
int DEPTH_ALWAYS;

// Culling constants
int CULL_BACK;
int CULL_FRONT;

//==============================================================================
//  Renderer::Renderer()
//==============================================================================
Renderer::Renderer() : m_frame(0) {
    m_current_camera = m_default_camera = Factory_camera.create( "camera" );
        
	m_shader_index = 0;
	m_cull_face = m_depth_func = m_mask = -1;
	m_blend_src = m_blend_dest = m_blend_mode = -1;    
}

//==============================================================================
//  Renderer::start_renderer()
//==============================================================================
void Renderer::start_renderer() {
    m_frame = 0;
    m_totaltime = 0.0f;
    m_starttime = base->get_sys_time();
    log_debug << std::endl << "Start rendering" << std::endl;
}

//==============================================================================
//  Renderer::stop_renderer()
//==============================================================================
void Renderer::stop_renderer() {
    if(m_totaltime == 0.0f) {
        m_totaltime = base->get_sys_time() - m_starttime;
    }
    
    log_debug << std::endl << "Rendering finished" << std::endl;
    log_debug << "  " << (m_frame) << " frame(s) rendered in " << m_totaltime << " seconds" << std::endl;
    log_debug << "  avarage: " << m_totaltime/(float)(m_frame) << ", fps: " << (float)(m_frame)/m_totaltime << std::endl;
}

//==============================================================================
//  Renderer::init()
//==============================================================================
void Renderer::init(Data_node* data) {
}

//==============================================================================
//  Renderer::add_3d_debug_line()
//==============================================================================
void Renderer::add_3d_debug_line( Vector3f start, Vector3f end ) { 
	m_debug_lines_3d.push_back(start);
	m_debug_lines_3d.push_back(end);
}

//==============================================================================
//  Renderer::add_3d_aa_box()
//==============================================================================
void Renderer::add_3d_debug_aa_box( Vector3f tlf, Vector3f brb ) {
	add_3d_debug_line( Vector3f( tlf.x, tlf.y, tlf.z ), Vector3f( brb.x, tlf.y, tlf.z ) );
	add_3d_debug_line( Vector3f( brb.x, tlf.y, tlf.z ), Vector3f( brb.x, brb.y, tlf.z ) );
	add_3d_debug_line( Vector3f( brb.x, brb.y, tlf.z ), Vector3f( tlf.x, brb.y, tlf.z ) );
	add_3d_debug_line( Vector3f( tlf.x, brb.y, tlf.z ), Vector3f( tlf.x, tlf.y, tlf.z ) );			
	
	add_3d_debug_line( Vector3f( tlf.x, tlf.y, brb.z ), Vector3f( brb.x, tlf.y, brb.z ) );
	add_3d_debug_line( Vector3f( brb.x, tlf.y, brb.z ), Vector3f( brb.x, brb.y, brb.z ) );
	add_3d_debug_line( Vector3f( brb.x, brb.y, brb.z ), Vector3f( tlf.x, brb.y, brb.z ) );
	add_3d_debug_line( Vector3f( tlf.x, brb.y, brb.z ), Vector3f( tlf.x, tlf.y, brb.z ) );			
		
	add_3d_debug_line( Vector3f( tlf.x, tlf.y, tlf.z ), Vector3f( tlf.x, tlf.y, brb.z ) );
	add_3d_debug_line( Vector3f( brb.x, tlf.y, tlf.z ), Vector3f( brb.x, tlf.y, brb.z ) );
	add_3d_debug_line( Vector3f( brb.x, brb.y, tlf.z ), Vector3f( brb.x, brb.y, brb.z ) );
	add_3d_debug_line( Vector3f( tlf.x, brb.y, tlf.z ), Vector3f( tlf.x, brb.y, brb.z ) );				
}

//==============================================================================
//  Renderer::add_3d_aa_box()
//==============================================================================
void Renderer::add_3d_debug_aa_box( AA_box* aa_box ) {
	add_3d_debug_aa_box( aa_box->get_centre() + aa_box->get_diagonal(),
				         aa_box->get_centre() - aa_box->get_diagonal() );
}
        
//==============================================================================
//  Renderer::add_2d_debug_line()
//==============================================================================
void Renderer::add_2d_debug_line( Vector2f start, Vector2f end ) {
    m_debug_lines_2d.push_back(start);
	m_debug_lines_2d.push_back(end);
}

//==============================================================================
//  Renderer::add_2d_debug_quad()
//==============================================================================
void Renderer::add_2d_debug_quad( Vector2f tl, Vector2f br ) {
	add_2d_debug_line( tl, Vector2f( tl.x, br.y ) ); 
	add_2d_debug_line( tl, Vector2f( br.x, tl.y ) ); 
	add_2d_debug_line( br, Vector2f( tl.x, br.y ) ); 
	add_2d_debug_line( br, Vector2f( br.x, tl.y ) ); 			
}

//==============================================================================
//  Renderer::render_debug_lines()
//==============================================================================
void Renderer::render_debug_lines() {
	m_debug_lines_3d.clear();	
	m_debug_lines_2d.clear();	
}

//==============================================================================
//  Renderer::render_debug()
//==============================================================================
void Renderer::render_debug() {
	if( m_debug_lines_3d.size() || m_debug_lines_2d.size() ) {
		render_debug_lines();
	}
}
