//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include <GL/gl.h>
#include <GL/glu.h>

#include "renderer_opengl.hpp"
#include "../base.hpp"
#include "../scene/batch.hpp"
#include "../misc/log.hpp"
#include "gl_extensions.hpp"
#include "../material/image.hpp"
#include "../material/material.hpp"
#include "../material/shader.hpp"
#include "../scene/scene.hpp"
#include "../scene/primitive.hpp"
#include "../scene/camera.hpp"
#include "../misc/factory.hpp"

#define MAX_TEXTURE_SLOTS 8

int DEFAULT_ANISOTROPY;

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_renderer_opengl  = Factory_renderer.reg<Renderer_opengl>( "opengl" );

//==============================================================================
// 
//==============================================================================
GLenum getGLPrimitive(const Primitive_type primType){
    switch (primType){
    case PRIM_TRIANGLES:      return GL_TRIANGLES;
    case PRIM_QUADS:          return GL_QUADS;
    case PRIM_TRIANGLE_STRIP: return GL_TRIANGLE_STRIP;
    case PRIM_POINTS:         return GL_POINTS;    
    default:
        return 0;
    }
}

GLenum arrayType[] = { 
	  GL_VERTEX_ARRAY, 
	  GL_TEXTURE_COORD_ARRAY,	// normal array
	  GL_TEXTURE_COORD_ARRAY, 
	  GL_COLOR_ARRAY,
	  GL_TEXTURE_COORD_ARRAY,
	  GL_TEXTURE_COORD_ARRAY
};

//============================================================================//
//                                                                            //
//    Batch_opengl                                                            //
//                                                                            //
//============================================================================//

//==============================================================================
//  Batch_opengl::Batch_opengl()
//==============================================================================
Batch_opengl::Batch_opengl() : m_indexbuffer(0), m_vertexbuffer(0) { }

//==============================================================================
//  Batch_opengl::upload_batch()
//==============================================================================
bool Batch_opengl::upload_batch() {
    if (GL_ARB_vertex_buffer_object_supported) {
        if(!m_vertexbuffer) {
            glGenBuffersARB(1, &m_vertexbuffer);
        }    
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexbuffer);
        glBufferDataARB(GL_ARRAY_BUFFER_ARB, m_num_vertices * m_vertex_size, m_vertices, GL_STATIC_DRAW_ARB);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

        if (m_num_indices > 0) {
		   	if(!m_indexbuffer) {
                glGenBuffersARB(1, &m_indexbuffer);
			}    
            glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexbuffer);
            glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_num_indices * sizeof(unsigned short), m_indices, GL_STATIC_DRAW_ARB);
            glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
        }
    }    
    return true;
}

//==============================================================================
//  Batch_opengl::render_solid()
//==============================================================================
void Batch_opengl::render_solid( int mask ) {
        unsigned int i, num_formats = m_formats.size();
        
        if(!m_vertexbuffer && GL_ARB_vertex_buffer_object_supported){
              upload_batch();
        }  
        if (GL_ARB_vertex_buffer_object_supported && m_vertexbuffer != 0) {
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexbuffer);
        }
       
        for (i=0; i<num_formats; ++i) {
            Format format = m_formats[i];
            GLenum array = arrayType[format.att_type];

            float *basePointer = (m_vertexbuffer)? NULL : get_vertices();
            
            switch (format.att_type) {
                case ATT_VERTEX:
                    glVertexPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
					glEnableClientState(array);
                    break;
            }
        }

        if (m_indexbuffer) {
              glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexbuffer);
              glDrawElements(getGLPrimitive( get_primitive_type() ), get_index_count(), GL_UNSIGNED_SHORT, NULL);
              glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
        } else {
              glDrawElements(getGLPrimitive( get_primitive_type() ), get_index_count(), GL_UNSIGNED_SHORT, get_indices());              
        }            

        for (i = 0; i<num_formats; ++i) {
            if (get_format(i).att_type == ATT_TEXCOORD) {
			    glDisableClientState(arrayType[get_format(i).att_type]);
			}
        }
        
    if ( GL_ARB_vertex_buffer_object_supported && m_vertexbuffer) {
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    }	 
}

//==============================================================================
//  Batch_opengl::render()
//==============================================================================
void Batch_opengl::render() {
        unsigned int i, num_formats = m_formats.size();
        
        if(!m_vertexbuffer && GL_ARB_vertex_buffer_object_supported){
              upload_batch();
        }      
  
        if (GL_ARB_vertex_buffer_object_supported && m_vertexbuffer != 0) {
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexbuffer);
        }
       
        for (i=0; i<num_formats; ++i) {
            Format format = m_formats[i];
            GLenum array = arrayType[format.att_type];

            float *basePointer = (m_vertexbuffer)? NULL : get_vertices();

            switch (format.att_type) {
                case ATT_TANGENT_S:
                    glClientActiveTextureARB(GL_TEXTURE2_ARB);
                    glTexCoordPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
                    break;
                case ATT_TANGENT_T:
                    glClientActiveTextureARB(GL_TEXTURE3_ARB);
                    glTexCoordPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
                    break;
                case ATT_VERTEX:
                    glVertexPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
                    break;
                case ATT_NORMAL:
                    glClientActiveTextureARB(GL_TEXTURE1_ARB);
                    glTexCoordPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
//                    glNormalPointer(GL_FLOAT, get_vertex_size(), basePointer + format.offset);
                    break;
                case ATT_TEXCOORD:
				    glClientActiveTextureARB(GL_TEXTURE0_ARB + format.index);
                    glTexCoordPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
                    break;
                case ATT_COLOR:
                    glColorPointer(format.size, GL_FLOAT, get_vertex_size(), basePointer + format.offset);
                break;
            }
            glEnableClientState(array);
        }

        if (m_indexbuffer) {
              glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexbuffer);
              glDrawElements(getGLPrimitive( get_primitive_type() ), get_index_count(), GL_UNSIGNED_SHORT, NULL);
              glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
        } else {
              glDrawElements(getGLPrimitive( get_primitive_type() ), get_index_count(), GL_UNSIGNED_SHORT, get_indices());              
        }            

        for (i = 0; i<num_formats; ++i) {
            if (get_format(i).att_type == ATT_TEXCOORD) {
                glClientActiveTextureARB(GL_TEXTURE0_ARB + get_format(i).index);
//              glClientActiveTextureARB(GL_TEXTURE0_ARB);            
            } else if(get_format(i).att_type == ATT_NORMAL) {
			    glClientActiveTextureARB(GL_TEXTURE1_ARB);
			} else if(get_format(i).att_type == ATT_TANGENT_S) {
			    glClientActiveTextureARB(GL_TEXTURE2_ARB);
			} else if(get_format(i).att_type == ATT_TANGENT_T) {
			    glClientActiveTextureARB(GL_TEXTURE3_ARB);
			}
            glDisableClientState(arrayType[get_format(i).att_type]);
        }
        
    if ( GL_ARB_vertex_buffer_object_supported && m_vertexbuffer) {
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    }
}

//============================================================================//
//                                                                            //
//    Texture_opengl                                                          //
//                                                                            //
//============================================================================//

//==============================================================================
//  Texture_opengl::init()
//==============================================================================
void Texture_opengl::init(Data_node* data) {
	Texture::init( data );
	
	m_anisotropy = DEFAULT_ANISOTROPY;
	data->get_value( "anisotropy", m_anisotropy );
}

//==============================================================================
//  Texture_opengl::upload_texture()
//==============================================================================
bool Texture_opengl::upload_texture() {
    log_function_multiple << "Upload texture" << std::endl;
    
    if(!m_images.size()) {
        log_debug << "No images in texture" << std::endl;
        log_error << "Texture binded unsuccessful" << std::endl;
        return false;
    }

    bool created = false;

    if(!m_tex_id) {
       glGenTextures( 1, &m_tex_id );
    }    
    if(m_tex_id) {
       created = true;
    }
 
    GLuint format, type, dimension;

    switch(m_images[0]->get_image_type() ) {
        case Image::RGB:       format = GL_RGB;
                               break;
        case Image::RGBA:      format = GL_RGBA;
                               break;
        case Image::GREYSCALE: format = GL_ALPHA;
                               break;
    }
    
    switch(m_type) {
        case BYTE:  type = GL_UNSIGNED_BYTE;
                    break;
        case FLOAT: type = GL_FLOAT;
                    log_error << "Float type image not supported!" << std::endl;
                    break;
    }
           
    switch(m_dimension) {
        case TEX_1D:   dimension = GL_TEXTURE_1D;
                       log_debug_multiple << "type = GL_TEXTURE_1D" << std::endl;
                       break;
        case TEX_2D:   dimension = GL_TEXTURE_2D;
                       log_debug_multiple << "type = GL_TEXTURE_2D" << std::endl;
                       break;
//        case TEX_3D:   dimension = GL_TEXTURE_3D;
//                       log_debug_multiple << "type = GL_TEXTURE_3D" << std::endl;
//                       break;
        case TEX_CUBE: if(!GL_ARB_texture_cube_map_supported) {
                           log_warning << "GL_TEXTURE_CUBE_MAP is not supported!" << std::endl;
                           dimension = GL_TEXTURE_2D;
                       } else {
                           dimension = GL_TEXTURE_CUBE_MAP_ARB;
                           log_debug_multiple << "type = GL_TEXTURE_CUBE_MAP_ARB" << std::endl;
                       }
                       break;
    }

    glBindTexture( dimension, m_tex_id);
        
    if(m_dimension != TEX_CUBE) {
        gluBuild2DMipmaps( dimension, format, m_images[0]->get_width(), m_images[0]->get_height(),
                             format, type, m_images[0]->get_pixels());
    } else {
        if(m_images.size()<6) {
            log_error << "Trying to upload Cube Map with " << m_images.size() << " images" << std::endl;
            return false;
        }
        
        GLuint cube_map_faces[] = {
            GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
            GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
            GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
            GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
            GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
            GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
        };
            
        for( int i=0; i<6; i++ ) {
            switch(m_images[i]->get_image_type() ) {
                case Image::RGB:       format = GL_RGB;
                                       break;
                case Image::RGBA:       format = GL_RGBA;
                                       break;
                case Image::GREYSCALE: format = GL_ALPHA;
                                          break;
            }
            gluBuild2DMipmaps( cube_map_faces[i], format, m_images[i]->get_width(), m_images[i]->get_height(),
                               format, type, m_images[i]->get_pixels() );
        }
    }

//        glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

    //Set Parameters
    if(!m_use_nearest) {
         glTexParameteri(dimension, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
         glTexParameteri(dimension, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    } else {
         glTexParameteri(dimension, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
         glTexParameteri(dimension, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
     }
        
    if(!m_use_clamp) {
        glTexParameteri(dimension, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(dimension, GL_TEXTURE_WRAP_T, GL_REPEAT);
        if(dimension == TEX_CUBE) {
            glTexParameteri(dimension, GL_TEXTURE_WRAP_R, GL_REPEAT);
        }
    } else {
        glTexParameteri(dimension, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(dimension, GL_TEXTURE_WRAP_T, GL_CLAMP);
        if(dimension == TEX_CUBE) {
            glTexParameteri(dimension, GL_TEXTURE_WRAP_R, GL_CLAMP);
        }
    }
   
    if (m_anisotropy && GL_EXT_texture_filter_anisotropic_supported){
            int max_anisotropy = 1;
            glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &max_anisotropy);
            if( max_anisotropy < m_anisotropy ) {
				m_anisotropy = max_anisotropy;
			}
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, m_anisotropy);
    }
               
    if(created) {
        log_debug_multiple << "width " << m_images[0]->get_width() << ", height " << m_images[0]->get_height() << " bind to " << m_tex_id << std::endl;
        log_debug_multiple << "mipmap = " << (m_use_nearest?"GL_NEAREST":"GL_LINEAR_MIPMAP_LINEAR") << std::endl;
        log_success_multiple << "Texture binded successful" << std::endl;
    } else {
        log_debug_multiple << "glGenTextures == NULL" << std::endl;
        log_error << "Texture uploaded unsuccessful" << std::endl;
    }
    
    return created;
}

//============================================================================//
//                                                                            //
//    Shader_opengl                                                           //
//                                                                            //
//============================================================================//

//==============================================================================
//  Shader_opengl::init() 
//==============================================================================
void Shader_opengl::init(Data_node* data) {
	Shader::init( data ); 
	 
	std::string s;
	
	if( data->get_value("file_glsl", s) ) {
		read_from_file( s );
 	} 
}

//==============================================================================
//  Shader_opengl::bind_value()
//==============================================================================
int Shader_opengl::bind_value( Shader_value& value ) {
	int index = glGetUniformLocationARB( m_index, value.get_name().c_str() );
	return index;
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, float constant) {
    if (index >= 0) glUniform1fARB(index, constant);	 
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, int constant) {
    if (index >= 0) glUniform1iARB(index, constant);	 
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, Vector2f &constant) {
	if (index >= 0) glUniform2fvARB(index, 1, (float *) constant);
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, Vector3f &constant) {
	if (index >= 0) glUniform3fvARB(index, 1, (float *) constant);	 
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, Vector4f &constant) {
	if (index >= 0) glUniform4fvARB(index, 1, (float *) constant);	 
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, Matrix3x3f &constant) {
	if (index >= 0) glUniformMatrix3fvARB(index, 1, GL_TRUE, (float *) &constant);
}

//==============================================================================
//  Shader_opengl::set_shader_constant()
//==============================================================================
void Shader_opengl::set_shader_constant(int index, Matrix4x4f &constant) {
	if (index >= 0) glUniformMatrix4fvARB(index, 1, GL_TRUE, (float *) &constant);	 
}

//==============================================================================
//  Shader_opengl::upload_shader()
//==============================================================================
void Shader_opengl::upload_shader(const std::string &vertex_shader_text, const std::string &fragment_shader_text) {
	GLint v_result, f_result, l_result;
	char str[4096];
	const char *shader_source;
	
	m_index = glCreateProgramObjectARB();
	log_debug << "Bind shader to " << m_index << std::endl;

	m_vertex_shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);

	shader_source = (char *)vertex_shader_text.c_str();
	glShaderSourceARB(m_vertex_shader, 1, &shader_source, NULL);
	glCompileShaderARB(m_vertex_shader);
	glGetObjectParameterivARB(m_vertex_shader, GL_OBJECT_COMPILE_STATUS_ARB, &v_result);
	if (v_result){
		glAttachObjectARB(m_index, m_vertex_shader);
	} else {
		log_warning << "Vertex shader error:" << std::endl;
		glGetInfoLogARB(m_vertex_shader, sizeof(str), NULL, str);
		log_warning << str << std::endl;
	}
	m_fragment_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

	shader_source = (char *)fragment_shader_text.c_str();
	glShaderSourceARB(m_fragment_shader, 1, &shader_source, NULL);
	glCompileShaderARB(m_fragment_shader);
	glGetObjectParameterivARB(m_fragment_shader, GL_OBJECT_COMPILE_STATUS_ARB, &f_result);
	if (f_result){
		glAttachObjectARB(m_index, m_fragment_shader);
	} else {
		log_warning << "Fragment shader error:" << std::endl;
		glGetInfoLogARB(m_fragment_shader, sizeof(str), NULL, str);
		log_warning << str << std::endl;
	}

	if (v_result && f_result){
		glLinkProgramARB(m_index);
		glGetObjectParameterivARB(m_index, GL_OBJECT_LINK_STATUS_ARB, &l_result);
		glGetInfoLogARB(m_index, sizeof(str), NULL, str);
		log_debug << str << std::endl;
	}
}

//============================================================================//
//                                                                            //
//    Renderer_opengl                                                         //
//                                                                            //
//============================================================================//

//==============================================================================
//  Renderer_opengl::Renderer_opengl()
//==============================================================================
Renderer_opengl::Renderer_opengl() {
    Factory_shader.reg< Shader_opengl >( "shader" );
    Factory_batch.reg< Batch_opengl >( "batch" );
	Factory_texture.reg< Texture_opengl >( "texture_2d" ); 
	Factory_texture.reg< Texture_opengl >( "texture_cube" ); 	

// Mask constants	
	BUFFER_DEPTH			  = GL_DEPTH_BUFFER_BIT;
	BUFFER_COLOR 			  = GL_COLOR_BUFFER_BIT;

// Blending constants
    BLEND_ZERO				  = GL_ZERO;
    BLEND_ONE				  = GL_ONE;
    BLEND_SRC_COLOR			  = GL_SRC_COLOR;
    BLEND_ONE_MINUS_SRC_COLOR = GL_ONE_MINUS_SRC_COLOR;
    BLEND_DST_COLOR 		  = GL_DST_COLOR;
    BLEND_ONE_MINUS_DST_COLOR = GL_ONE_MINUS_DST_COLOR;
    BLEND_SRC_ALPHA			  = GL_SRC_ALPHA;
    BLEND_ONE_MINUS_SRC_ALPHA = GL_ONE_MINUS_SRC_ALPHA;
    BLEND_DST_ALPHA			  = GL_DST_ALPHA;
    BLEND_ONE_MINUS_DST_ALPHA = GL_ONE_MINUS_DST_ALPHA;
    BLEND_SRC_ALPHA_SATURATE  = GL_SRC_ALPHA_SATURATE;

    BLEND_ADD				  = GL_FUNC_ADD_EXT;
    BLEND_SUBTRACT			  = GL_FUNC_SUBTRACT_EXT;
    BLEND_REVERSE_SUBTRACT 	  = GL_FUNC_REVERSE_SUBTRACT_EXT;
    BLEND_MIN 				  = GL_MIN_EXT;
    BLEND_MAX				  = GL_MAX_EXT;

// Depth testing constants
    DEPTH_NEVER 			  = GL_NEVER;
    DEPTH_LESS				  = GL_LESS;
    DEPTH_EQUAL				  = GL_EQUAL;
    DEPTH_LEQUAL			  = GL_LEQUAL;
    DEPTH_GREATER			  = GL_GREATER;
    DEPTH_NOTEQUAL			  = GL_NOTEQUAL;
    DEPTH_GEQUAL			  = GL_GEQUAL;
    DEPTH_ALWAYS			  = GL_ALWAYS;

// Culling constants
    CULL_BACK				  = GL_BACK;
    CULL_FRONT				  = GL_FRONT;
	
	DEFAULT_ANISOTROPY = 0;
	
	m_active_texture_slots.resize( MAX_TEXTURE_SLOTS );
	for(int i=0; i<MAX_TEXTURE_SLOTS; ++i) {
		m_active_texture_slots[i] = -1;	
	}
}

//==============================================================================
//  Renderer_opengl::init()
//==============================================================================
void Renderer_opengl::init(Data_node* data) {
	Renderer::init( data );
	
	DEFAULT_ANISOTROPY = 0;
	data->get_value( "default_anisotropy", DEFAULT_ANISOTROPY);
}

//==============================================================================
//  Renderer_opengl::init()
//==============================================================================
void Renderer_opengl::init() {
    initExtensions();
 
	set_blending( BLEND_NONE, BLEND_NONE );
	set_depth_func( DEPTH_LEQUAL );
	set_cull_face( CULL_BACK );
    
    glEnable(GL_SCISSOR_TEST);    

    glClearColor( 0, 0, 0, 1.0f);
}

//==============================================================================
//  Renderer_opengl::close()
//==============================================================================
void Renderer_opengl::close() {
}    

//==============================================================================
//  Renderer_opengl::set_viewport()
//==============================================================================
void Renderer_opengl::set_viewport(const int left, const int top, const int width, const int height) {
    glViewport(left, top, width, height);

    glMatrixMode(GL_PROJECTION);                       
    glLoadIdentity();

    gluPerspective(45.0f, (float)width/(float)height, 0.15f ,500000.0f);    
    set_renderport(left, top, width, height);
}

//==============================================================================
//  Renderer_opengl::set_renderport()
//==============================================================================
void Renderer_opengl::set_renderport(const int left, const int top, const int width, const int height) {
    set_scissor(left, top, width, height);
    
	m_left = left;
	m_top = top;
	m_width = width;
	m_height = height;
}    

//==============================================================================
//  Renderer_opengl::set_view_rotation()
//==============================================================================
void Renderer_opengl::set_view_rotation(Camera* camera) {
    set_camera( camera );
    
    glMatrixMode(GL_MODELVIEW); 
    glLoadMatrixf( Matrix4x4f( camera->get_orientation() ) ); 
}

//==============================================================================
//  Renderer_opengl::set_view()
//==============================================================================
void Renderer_opengl::set_view(Camera* camera) {
    set_camera( camera );
    
    glMatrixMode(GL_MODELVIEW); 
    glLoadMatrixf( camera->get_modelview_matrix() );
}

//==============================================================================
//  Renderer_opengl::bind_texture()
//==============================================================================
void Renderer_opengl::bind_texture(int texture_slot, Texture* texture) {
 	if( !texture->get_tex_id() ) {
	    ((Texture_opengl *)texture)->upload_texture();;
	}
	activate_texture_slot( texture_slot, texture->get_tex_id() );
}

//==============================================================================
//  Renderer_opengl::activate_texture_slot()
//==============================================================================
void Renderer_opengl::activate_texture_slot(int texture_slot, int tex_id) {
	if( m_active_texture_slots[ texture_slot ] < 0 || m_active_texture_slots[ texture_slot ] != tex_id ) {
		glActiveTextureARB(GL_TEXTURE0_ARB + texture_slot);		
		if( m_active_texture_slots[ texture_slot ] < 0 ) {
//			glEnable( GL_TEXTURE_2D );
		}		
		glBindTexture(GL_TEXTURE_2D, tex_id );		
	}
	m_active_texture_slots[ texture_slot ] = tex_id;
}

//==============================================================================
//  Renderer_opengl::disable_texture_slot()
//==============================================================================
void Renderer_opengl::disable_texture_slot(int texture_slot) {
	if( m_active_texture_slots[ texture_slot ] >= 0 ) {
		glActiveTextureARB(GL_TEXTURE0_ARB + texture_slot);
		glDisable( GL_TEXTURE_2D );
		m_active_texture_slots[ texture_slot ] = -1;
	} 
}

//==============================================================================
//  Renderer_opengl::disable_texture_slots()
//==============================================================================
void Renderer_opengl::disable_texture_slots() {
	for(int i=0; i<MAX_TEXTURE_SLOTS; ++i) {
		disable_texture_slot(i);	
	}	
}

//==============================================================================
//  Renderer_opengl::clear_buffer()
//==============================================================================
void Renderer_opengl::clear_buffer( int mask ) {
    glClear( mask );
}

//==============================================================================
//  Renderer_opengl::enable_2d_mode()
//==============================================================================
void Renderer_opengl::enable_2d_mode( int width, int height ) {
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho( 0, width, 0, height, 1, 10);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glTranslatef(0, 0, -5);
}

//==============================================================================
//  Renderer_opengl::disable_2d_mode()
//==============================================================================
void Renderer_opengl::disable_2d_mode() {
	glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);    
}

//==============================================================================
//  Renderer_opengl::render_scene()
//==============================================================================
void Renderer_opengl::render_scene( std::vector<Primitive *> *primitives, std::vector<Primitive *> *lights, Background* background) {
    set_default();
	set_color( black );
	
    for(int i=0, s=primitives->size(); i<s; ++i) {
		if(get_camera()->intersect( primitives->at(i)->get_bounding_box() ) ) {
			primitives->at(i)->get_batch()->render_solid( BUFFER_COLOR | BUFFER_DEPTH );
		}  
	}
	
	set_blending( BLEND_ONE, BLEND_ONE );

//	Shader* shader = Factory_shader.use( "diff_bump_spec_map" );
	Shader* shader = (*Factory_shader.get_objects().begin()).second;
	set_shader( shader );
	
	shader->set_shader_global_constants();
	shader->set_shader_constants();
	
	for(int j=0, l=lights->size(); j<l; ++j) {
		shader->set_shader_light_constants( lights->at(j) );
		for(int i=0, s=primitives->size(); i<s; ++i) {
			if( get_camera()->intersect(primitives->at(i)->get_bounding_box())) {
				shader->set_shader_material_constants( primitives->at(i)->get_material() );
				primitives->at(i)->get_batch()->render();
            }
        }
	}
}

//==============================================================================
//  Renderer_opengl::set_shader()
//==============================================================================
void Renderer_opengl::set_shader(Shader* shader) {
	unsigned int shader_index; 
	if (shader != SHADER_NONE){
	   	shader_index = shader->get_index();	   
	} else {
	  	shader_index = 0;   
    }    		   
	if (shader_index != m_shader_index){
		glUseProgramObjectARB(shader_index);
		m_shader_index = shader_index;
	}
}

//==============================================================================
//  Renderer_opengl::render_debug_lines()
//==============================================================================
void Renderer_opengl::render_debug_lines() {    
    set_scissor( 0, 0, base->get_window_width(), base->get_window_height() );    
	set_default();
	set_depth_func( DEPTH_NONE );
	
	// 3d debug lines
	
	set_color( green );
    glBegin(GL_LINES);
    for(int i=0, l=m_debug_lines_3d.size(); i<l; ++i) {
        glVertex3fv( m_debug_lines_3d[i] ); 
    }
    glEnd(); 

	// 2d debug lines

	enable_2d_mode( base->get_window_width(), base->get_window_height() );
	
	set_color( red );
    glBegin(GL_LINES);
    for(int i=0, l=m_debug_lines_2d.size(); i<l; ++i) {
        glVertex2fv( m_debug_lines_2d[i] ); 
    }
    glEnd();    

	disable_2d_mode();

	m_debug_lines_3d.clear();
	m_debug_lines_2d.clear();
}

//==============================================================================
//  Renderer_opengl::set_cull_face()
//==============================================================================
void Renderer_opengl::set_cull_face(const int cull_face) {
 	if(cull_face != m_cull_face) {
		if(cull_face == CULL_NONE) {
			glDisable(GL_CULL_FACE);
		} else {
			if (m_cull_face == CULL_NONE) glEnable(GL_CULL_FACE);
			glCullFace(cull_face);
		}
		m_cull_face = cull_face;
	}
}

//==============================================================================
//  Renderer_opengl::set_depth_func()
//==============================================================================
void Renderer_opengl::set_depth_func(const int depth_func) {
 	if (depth_func != m_depth_func) {
		if (depth_func == DEPTH_NONE) {
			glDisable(GL_DEPTH_TEST);
		} else {
			if(m_depth_func == DEPTH_NONE) {
				glEnable(GL_DEPTH_TEST);
			}
			glDepthFunc(depth_func);
		}
		m_depth_func = depth_func;
	}
}

//==============================================================================
//  Renderer_opengl::set_mask()
//==============================================================================
void Renderer_opengl::set_mask(const int mask) {
/*
	 	if (mask != m_mask){
		if ((mask & DEPTH) != (m_mask & DEPTH)){
			glDepthMask(mask & DEPTH);
		}
		if ((mask & COLOR) != (currentMask & COLOR)){
			glColorMask(((mask & RED) >> 1) & 1, ((mask & GREEN) >> 2) & 1, ((mask & BLUE) >> 3) & 1, ((mask & ALPHA) >> 4) & 1);
		}
		m_mask = mask;
	}
*/	
}

//==============================================================================
//  Renderer_opengl::set_blending()
//==============================================================================
void Renderer_opengl::set_blending(const int src, const int dest, const int mode) {
 	if (src != BLEND_NONE) {
		if (m_blend_src == BLEND_NONE) {
			glEnable(GL_BLEND);
		}
		if (src != m_blend_src || dest != m_blend_dest) {
			glBlendFunc(m_blend_src = src, m_blend_dest = dest);
		}
		if (mode != m_blend_mode) {
//			glBlendEquation(m_blend_mode = mode);
		}
	} else {
		if (m_blend_src != BLEND_NONE){
			glDisable(GL_BLEND);
			m_blend_src = BLEND_NONE;
		}
	}
}

//==============================================================================
//  Renderer_opengl::set_scissor()
//==============================================================================
void Renderer_opengl::set_scissor(const int x, const int y, const int w, const int h) {
	glScissor(x, y, w, h); 
}

//==============================================================================
//  Renderer_opengl::set_color()
//==============================================================================
void Renderer_opengl::set_color( const Color& color ) {
	glColor3fv( color ); 
}

//==============================================================================
//  Renderer_opengl::set_default()
//==============================================================================
void Renderer_opengl::set_default() {
	set_blending( BLEND_NONE, BLEND_NONE );
	set_depth_func( DEPTH_LEQUAL );
	disable_texture_slots();
	set_shader( SHADER_NONE );
}
