//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "batch.hpp"

#include <stdio.h>
#include <fstream>

#include "../math/vector.hpp"
#include "../misc/log.hpp"
#include "../misc/data.hpp"
#include "../misc/factory.hpp" 

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_batch = Factory_batch.reg< Batch >( "batch" );


//==============================================================================
//  set_value()
//==============================================================================
void set_value(const char *dest, const unsigned int index, float value) {
    *(((float *) dest) + index) = value;
}

//==============================================================================
//  Batch::Batch()
//==============================================================================
Batch::Batch() {
    m_vertices = NULL;
    m_indices  = NULL;

	clean();
}

//==============================================================================
//  Batch::~Batch()
//==============================================================================
Batch::~Batch() {
    clean();
}

//==============================================================================
//  Batch::clean()
//==============================================================================
void Batch::clean() {
    if (m_vertices != NULL) delete[] m_vertices;
    if (m_indices  != NULL) delete[] m_indices;
    m_vertices = NULL;
    m_indices = NULL;    
    
    m_num_vertices = 0;
    m_num_indices  = 0;
    m_vertex_size = 0;
    m_vertex_elements = 0;
    
    m_tangent_s_offset = m_tangent_t_offset = m_vertex_offset = m_normal_offset = m_texcoord_offset = 0;
    
    m_primitive_type = PRIM_TRIANGLES;
}

//==============================================================================
//  Batch::init()
//==============================================================================
void Batch::init(Data_node* data) {
	clean();
	
	std::string s;
    if( data->get_value( "file_bin", s ) ) {
        read_from_file_bin( s );
	}
	if( data->get_value( "file_mesh", s ) ) {
        read_from_file_mesh( s );
	} 
}
    
//==============================================================================
//  Batch::find_attribute()
//==============================================================================
bool Batch::find_attribute(const Attribute_type att_type, const unsigned int index, unsigned int *where) const {
    for (unsigned int i = 0; i < m_formats.size(); i++) {
        if (m_formats[i].att_type == att_type && m_formats[i].index == index) {
            if (where != NULL) *where = i;
            return true;
        }
    }
    return false;
}

//==============================================================================
//  Batch::insert_attribute()
//==============================================================================
bool Batch::insert_attribute(const Attribute_type att_type, const unsigned int size, const unsigned int index) {
    unsigned int i;

    for (i = 0; i < m_formats.size(); i++) {
        if (m_formats[i].att_type == att_type && m_formats[i].index == index) return false;
    }

    add_format(att_type, size, m_vertex_elements, index);

    unsigned int old_vertex_elements = m_vertex_elements - size;
    float *new_vertices = new float[m_num_vertices * m_vertex_elements];

    float *dest = new_vertices;
    float *src = m_vertices;
    for(i=0; i<m_num_vertices; ++i) {
        memcpy(dest, src, old_vertex_elements * sizeof(float) );
        memset(&dest[old_vertex_elements], 0, size * sizeof(float) );
        dest += m_vertex_elements;
        src += old_vertex_elements;
    }

    delete[] m_vertices;    
    m_vertices = new_vertices;
    
    return true;
}

//==============================================================================
//  Batch::remove_attribute()
//==============================================================================
bool Batch::remove_attribute(const Attribute_type att_type, const unsigned int index) {
    unsigned int i, j;

    for (i = 0; i < m_formats.size(); i++) {
        if (m_formats[i].att_type == att_type && m_formats[i].index == index) {
            Format format = m_formats[i];
            std::vector<Format>::iterator it = m_formats.begin() + i;
            m_formats.erase(it);

            for (j = i; j < m_formats.size(); j++) m_formats[i].offset -= format.size;

            unsigned int newm_vertex_size = m_vertex_size - format.size * sizeof(float);
            float *newm_vertices = new float[m_num_vertices * newm_vertex_size / sizeof(float) ];

            float *dest = newm_vertices;
            float *src = m_vertices;
            memcpy(dest, src, format.offset * sizeof( float ) );
            dest += format.offset * sizeof( float );
            src += (format.offset + format.size)  * sizeof( float );
            for (j = 0; j < m_num_vertices - 1; j++) {
                memcpy(dest, src, newm_vertex_size);
                dest += newm_vertex_size;
                src += m_vertex_size;
            }
            memcpy(dest, src, (m_vertex_size - (format.offset + format.size) * sizeof(float)));

            delete[] m_vertices;
            m_vertices = newm_vertices;
            m_vertex_size = newm_vertex_size;

            return true;
        }
    }
    return false;
}

//==============================================================================
//  Batch::add_normals()
//==============================================================================
bool Batch::add_normals() {
    if (m_indices == NULL) return false;
    if (m_primitive_type != PRIM_TRIANGLES) return false;    
    if (!find_attribute(ATT_NORMAL) && !insert_attribute(ATT_NORMAL, 3) ) return false;

	for(unsigned int i=0; i<m_num_indices; ++i) {
		get_normal_by_index( i ) = Vector3f( 0.0f, 0.0f, 0.0f );
    }
	for(unsigned int i=0; i<m_num_indices; i+=3) {
		Vector3f v1 = get_vertex_by_index( i+0 );
		Vector3f v2 = get_vertex_by_index( i+1 );
		Vector3f v3 = get_vertex_by_index( i+2 );
		Vector3f normal = cross( v2-v1, v3-v1 ).get_normalized();
		
		get_normal_by_index( i+0 ) += normal;
		get_normal_by_index( i+1 ) += normal;
		get_normal_by_index( i+2 ) += normal;
	}
	for(unsigned int i=0; i<m_num_vertices; ++i) {
		get_normal( i ).normalize();
    }

    return true;
}


//==============================================================================
//  Batch::add_tangent_space()
//==============================================================================
bool Batch::add_tangent_space() {
    if (m_indices == NULL) return false;
    if (!find_attribute(ATT_TEXCOORD) ) return false;    
    if (m_primitive_type != PRIM_TRIANGLES) return false;    
    if (!find_attribute(ATT_TANGENT_S) && !insert_attribute(ATT_TANGENT_S, 3) ) return false;
    if (!find_attribute(ATT_TANGENT_T) && !insert_attribute(ATT_TANGENT_T, 3) ) return false;

	for(unsigned int i=0; i<m_num_indices; ++i) {
		get_tangent_s_by_index( i ) = Vector3f( 0.0f, 0.0f, 0.0f );
		get_tangent_t_by_index( i ) = Vector3f( 0.0f, 0.0f, 0.0f );		
    }

	for(unsigned int i=0; i<m_num_indices; i+=3) {
		const Vector3f v1 = get_vertex_by_index( i+0 );
		const Vector3f v2 = get_vertex_by_index( i+1 );
		const Vector3f v3 = get_vertex_by_index( i+2 );
		const Vector2f w1 = get_texcoord_by_index( i+0 );
		const Vector2f w2 = get_texcoord_by_index( i+1 );
		const Vector2f w3 = get_texcoord_by_index( i+2 );
		
		float x1 = v2.x - v1.x;
		float x2 = v3.x - v1.x;
		float y1 = v2.y - v1.y;
		float y2 = v3.y - v1.y;
		float z1 = v2.z - v1.z;
		float z2 = v3.z - v1.z;
		
		float s1 = w2.x - w1.x;
		float s2 = w3.x - w1.x;
		float t1 = w2.y - w1.y;
		float t2 = w3.y - w1.y;
		
		float r;
		
		if(fabs(s1*t2 - s2*t1) <= 0.0001f) {
			r = 1.0f;
  		} else {
			r = 1.0f / (s1*t2 - s2*t1);
		}
		Vector3f sdir((t2 * x1 - t1 * x2) * r, (t2 * y1 - t1 * y2) * r, (t2 * z1 - t1 * z2) * r);
		Vector3f tdir((s1 * x2 - s2 * x1) * r, (s1 * y2 - s2 * y1) * r, (s1 * z2 - s2 * z1) * r);
			
		get_tangent_s_by_index( i+0 ) += sdir;
		get_tangent_s_by_index( i+1 ) += sdir;
		get_tangent_s_by_index( i+2 ) += sdir;
		
		get_tangent_t_by_index( i+0 ) += tdir;
		get_tangent_t_by_index( i+1 ) += tdir;
		get_tangent_t_by_index( i+2 ) += tdir;		
	}
	
	for(unsigned int i=0; i<m_num_vertices; ++i) {
		const Vector3f n = get_normal(i);
		
		// Gram-Schmidt orthogonalize
		get_tangent_t(i) = (get_tangent_t(i) - n * dot(n, get_tangent_t(i))).get_normalized();
		get_tangent_s(i) = (get_tangent_s(i) - n * dot(n, get_tangent_s(i)) - n * dot(n, get_tangent_t(i))).get_normalized();
    }

    return true;   
}



	
//==============================================================================
//  Batch::read_from_file_mesh()
//==============================================================================
void Batch::read_from_file_mesh(const std::string& name) {
    clean();
    
	std::ifstream file( name.c_str() );
    if(!file) {
        return;
    } 
    
    std::string s;

	while( file >> s ) {
		if( s == "format" ) {
		    int size, offset, index = 0;
		    Attribute_type type = ATT_COLOR;
						   
			file >> s >> size >> offset >> index;
			if( s == "vertex" ) {
				type = ATT_VERTEX;
			} else if( s == "normal" ) {
				type = ATT_NORMAL;
			} else if( s == "texcoord" ) {
				type = ATT_TEXCOORD;
   			} else if( s == "tangent_s" ) {
				type = ATT_TANGENT_S;
   			} else if( s == "tangent_t" ) {
				type = ATT_TANGENT_T;
   			} 
            add_format( type, size, offset, index); 
		} else if( s == "num_vertices" ) {
			file >> m_num_vertices;
		} else if ( s == "num_indices" ) {
		    file >> m_num_indices;   
        } else if ( s == "vertex_elements" ) {
		  	file >> m_vertex_elements;		  	
  	    } else if ( s == "vertices" ) {	  	
		  	m_vertex_size = m_vertex_elements * sizeof(float);
		    m_vertices = new float[m_num_vertices * m_vertex_elements];	
		  	for(int i=0; i<m_num_vertices * m_vertex_elements; ++i) {
				file >> m_vertices[i];	
		    }   
  	    } else if ( s == "indices" ) {
	        m_indices = new unsigned short[m_num_indices];
		  	for(int i=0; i<m_num_indices; ++i) {
				file >> m_indices[i];	
		    }   
  	    }		  	   
    }

    if( m_num_indices == 0 ) {
		m_num_indices = m_num_vertices;
		
        m_indices = new unsigned short[m_num_indices];
        for(unsigned int i=0; i<m_num_indices; ++i) {
			m_indices[i] = i;	
		}
	}	
/*
    fread(&m_primitive_type, sizeof(m_primitive_type), 1, file);
*/
    add_tangent_space();
}

//==============================================================================
//  Batch::write_to_file_mesh()
//==============================================================================
void Batch::write_to_file_mesh(const std::string& name) {
								
	std::ofstream file( name.c_str() );
    if(!file) {
        return;
    } 
    
    for(unsigned int i=0; i<m_formats.size(); ++i) {
        file << "format\t\t";
        Attribute_type type = m_formats[i].att_type;
        switch( type ) {
			case ATT_VERTEX:    file << "vertex   ";
			                    break; 	
			case ATT_NORMAL:    file << "normal   ";
			                    break;	
			case ATT_TEXCOORD:  file << "texcoord ";
			                    break;	
			case ATT_COLOR:     file << "color    ";
			                    break;	
			case ATT_TANGENT_S: file << "tangent_s";
			                    break;	
			case ATT_TANGENT_T: file << "tangent_t";
			                    break;								   			                   
	    }
	    file << " " << m_formats[i].size << " " << m_formats[i].offset << " " << m_formats[i].index << std::endl;
	}
	file << std::endl;
	file << "num_vertices\t" << m_num_vertices << std::endl;
	file << "num_indices\t\t" << m_num_indices << std::endl;
	file << "vertex_elements\t" << m_vertex_elements << std::endl << std::endl;

	file << "vertices" << std::endl;
	for(unsigned int i=0; i<m_num_vertices * m_vertex_elements; ++i) {
	    file << m_vertices[i] << " ";
		if( !((i+1) % m_vertex_elements) ) file << std::endl;
    }
    file << std::endl << "indices" << std::endl;
    for(unsigned int i=0; i<m_num_indices; ++i) {
	    file << m_indices[i] << " ";
		if( !((i+1) % 3) ) file << std::endl;
    }
    file.close();
}

//==============================================================================
//  Batch::read_from_file_bin()
//==============================================================================
void Batch::read_from_file_bin(const std::string& name) {
	clean();
	 
    FILE *file = fopen(name.c_str(), "rb");
    if (file == NULL) return;

    fread(&m_num_vertices, sizeof(m_num_vertices), 1, file);
    fread(&m_num_indices,  sizeof(m_num_indices),  1, file);
    fread(&m_vertex_size,  sizeof(m_vertex_size),  1, file);

    fread(&m_primitive_type, sizeof(m_primitive_type), 1, file);
   
    unsigned int num_of_formats;
    fread(&num_of_formats, sizeof(num_of_formats), 1, file);
    m_formats.resize(num_of_formats);
    fread(&m_formats.at(0), num_of_formats * sizeof(Format), 1, file);

    m_vertex_elements = m_vertex_size / sizeof(float);

    m_vertices = new float[m_num_vertices * m_vertex_elements];
    fread(m_vertices, m_vertex_size, m_num_vertices, file);

    if (m_num_indices > 0) {
        m_indices = new unsigned short[m_num_indices];
        fread(m_indices, sizeof(unsigned short), m_num_indices, file);
    } else {
	    m_indices = NULL;
    }
        
    fclose(file);    
}

//==============================================================================
//  Batch::write_to_file_bin()
//==============================================================================
void Batch::write_to_file_bin(const std::string& name) {
    FILE *file = fopen(name.c_str(), "wb");
    if (file == NULL) return;

    fwrite(&m_num_vertices, sizeof(m_num_vertices), 1, file);
    fwrite(&m_num_indices,  sizeof(m_num_indices),  1, file);
    fwrite(&m_vertex_size,  sizeof(m_vertex_size),  1, file);

    fwrite(&m_primitive_type, sizeof(m_primitive_type), 1, file);

    unsigned int num_formats = m_formats.size();
    fwrite(&num_formats, sizeof(num_formats), 1, file);
    fwrite(&m_formats.at(0), num_formats * sizeof(Format), 1, file);

    fwrite(m_vertices, m_vertex_size, m_num_vertices, file);
    fwrite(m_indices,  sizeof(unsigned short), m_num_indices, file);
        
    fclose(file);    
}

