#include "camera.hpp"

#include "../misc/log.hpp"
#include "../misc/data.hpp"
#include "../misc/factory.hpp"

#include "../renderer.hpp"
#include "../base.hpp"

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_camera  = Factory_camera.reg<Camera>( "camera" );
  
//==============================================================================
//  Camera::Camera()
//==============================================================================
Camera::Camera() : Controllable_object() {
    m_exposure = 0.0f;
    m_type = CAM_3D; 
 
    m_viewport_width  = (float)base->get_window_width();
    m_viewport_height = (float)base->get_window_height();  
	 
    set_perspective(45.0f, 0.15f ,500000.0f);  
}

//==============================================================================
//  Camera::init()
//==============================================================================
void Camera::init(Data_node* data) {
	Controllable_object::init( data );
	 
    Vector3f up, center = Vector3f(0.0f, 0.0f, 1.0f);
    
    bool use_camera = false;
    
    data->get_value( "use_camera", use_camera );
    data->get_value( "exposure", m_exposure );
    data->get_value( "view", center );
    data->get_value( "up", up );

/*  if( s == "perspective" ) {
           in->read( m_fov_y );
           in->read( m_near_plane );
           in->read( m_far_plane );*/

    if( use_camera ) renderer->set_camera( this );
        
	get_controller()->look_at( get_position(), center, up );
}

//==============================================================================
//  Camera::set_perspective()
//==============================================================================
void Camera::set_perspective(float fov_y, float near_plane, float far_plane) {
    m_fov_y = fov_y;
    m_near_plane = near_plane;
    m_far_plane = far_plane;
}

//==============================================================================
//  Camera::get_perspective()
//==============================================================================
void Camera::get_perspective(float &fov_y, float &near_plane, float &far_plane) {
    fov_y = m_fov_y;
    near_plane = m_near_plane;
    far_plane = m_far_plane;
}

//==============================================================================
//  Camera::update()
//==============================================================================
void Camera::update() {
    Controllable_object::update();
     
    // update matrices
    m_modelview_matrix = Matrix4x4f( get_orientation() );    
    Vector3f position = get_position();

    Matrix4x4f translation;
    translation.load_identity();
    translation.m_data[12] = -position.x;
    translation.m_data[13] = -position.y;
    translation.m_data[14] = -position.z;

    m_modelview_matrix = m_modelview_matrix * translation;
     
    if(m_type == CAM_3D) {
        m_projection_matrix.set_perspective( m_fov_y, m_viewport_width/m_viewport_height, m_near_plane, m_far_plane );
    } else {
        m_projection_matrix.set_ortho( 0, m_viewport_width, m_viewport_height, 0, m_near_plane, m_far_plane);
    }            
    m_projection_modelview_matrix = (m_projection_matrix * m_modelview_matrix);
    
    // calculate some variables for get_ray ...
    
	float radians = m_fov_y / 2.0f * M_PI / 180.0f;
	float maxy = tanf( radians ) ;
	float maxx = maxy / m_viewport_height * m_viewport_width;
	
	Vector3f right      = maxx * get_orientation().get_column( 0 );
	Vector3f up         = maxy * get_orientation().get_column( 1 );

	m_direction  = -get_orientation().get_column( 2 );
	
	m_direction -= right;
	m_direction += up;
		
	m_dx = right * (2.0f/m_viewport_width);
	m_dy = -up * (2.0f/m_viewport_height);
    	    
    update_frustum();
}

//==============================================================================
//  Camera::update_frustum()
//==============================================================================
void Camera::update_frustum() {
    const Matrix4x4f &clip = m_projection_modelview_matrix;

    m_frustum_planes[FRUSTUM_RIGHT ] = Plane(
    Vector3f( clip[ 3] - clip[ 0], clip[ 7] - clip[ 4], clip[11] - clip[ 8]), (clip[15] - clip[12]));

    m_frustum_planes[FRUSTUM_LEFT  ] = Plane(
    Vector3f( clip[ 3] + clip[ 0], clip[ 7] + clip[ 4], clip[11] + clip[ 8]), (clip[15] + clip[12]));

    m_frustum_planes[FRUSTUM_BOTTOM] = Plane(
    Vector3f( clip[ 3] + clip[ 1], clip[ 7] + clip[ 5], clip[11] + clip[ 9]), (clip[15] + clip[13]));

    m_frustum_planes[FRUSTUM_TOP   ] = Plane(
    Vector3f( clip[ 3] - clip[ 1], clip[ 7] - clip[ 5], clip[11] - clip[ 9]), (clip[15] - clip[13]));

    m_frustum_planes[FRUSTUM_BACK  ] = Plane(    
    Vector3f( clip[ 3] - clip[ 2], clip[ 7] - clip[ 6], clip[11] - clip[10]), (clip[15] - clip[14]));

    m_frustum_planes[FRUSTUM_FRONT ] = Plane(    
    Vector3f( clip[ 3] + clip[ 2], clip[ 7] + clip[ 6], clip[11] + clip[10]), (clip[15] + clip[14]));

    for(int i=0; i<FRUSTUM_PLANES; ++i) {
        m_frustum_planes[i].normalize();
    }        
} 
