//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "primitive.hpp"
#include "../renderer.hpp"
#include "../material/material.hpp"
#include "../math/geometry.hpp"
#include "../misc/log.hpp"

#include "../misc/factory.hpp"
#include "batch.hpp"

#include <sstream>
#include <iostream>
#include <fstream>

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_point = Factory_primitive.reg<Primitive_point>( "point" );
bool fac_sphere = Factory_primitive.reg<Primitive_sphere>( "sphere" );
bool fac_plane  = Factory_primitive.reg<Primitive_plane>( "plane" );
bool fac_mesh   = Factory_primitive.reg<Primitive_mesh>( "mesh" );

//==============================================================================
//  Primitive::Primitive()
//==============================================================================
Primitive::Primitive() {
//    m_material = Factory_material.use( "default" );
    set_type( Primitive::PRIMITIVE );
    set_fall_off( Primitive::QUADRATIC );
    m_bounding_box = new AA_box();
    m_cast_shadow = true;
    m_batch = NULL;
	m_intensity = 0.0f;
    
    m_max_samples = 64;
}

//==============================================================================
//  Primitive::~Primitive()
//==============================================================================
Primitive::~Primitive() {
     if( m_bounding_box ) delete m_bounding_box;
}

//==============================================================================
//  Primitive::init()
//==============================================================================
void Primitive::init(Data_node* data) {     
    Controllable_object::init( data );

	std::string s;

    if( Data_value* value = data->get_value( "material" ) ) {        
        m_material = Factory_material.create( value );
    }
    data->get_value( "cast_shadow", m_cast_shadow );
    data->get_value( "intensity", m_intensity );
    data->get_value( "max_samples", m_max_samples );   
    if( data->get_value( "fall_off", s ) ) {
        if( s == "linear" ) {
            m_fall_off = Primitive::LINEAR;
        } else if ( s == "quadratic" ) {
            m_fall_off = Primitive::QUADRATIC;
        } else {
            m_fall_off = Primitive::NONE;
        }
    }
	if( data->get_value( "type", s ) ) {
        if( s == "light" ) {
			set_type( Primitive::LIGHT );
		}
	}
}

//==============================================================================
//  Primitive_point::init()
//==============================================================================
void Primitive_point::init(Data_node* data) {     
    Primitive::init( data );

    m_max_samples = 1;
}

//==============================================================================
//  Primitive_point::calculate_bounding_box()
//==============================================================================    
void Primitive_point::calculate_bounding_box() {
	m_bounding_box->set( get_position(), Vector3f(0,0,0) );	
}

//==============================================================================
//  Primitive_sphere::init()
//==============================================================================
void Primitive_sphere::init(Data_node* data) {
    Primitive::init( data );
 
	m_tex_u = Vector3f( 1.0f, 0.0f, 0.0f );
	m_tex_v = Vector3f( 0.0f, 1.0f, 0.0f );
	m_tex_w = Vector3f( 0.0f, 0.0f, 1.0f );
	    
    data->get_value( "centre", get_position() );  
    data->get_value( "radius", m_radius );
    
    data->get_value( "tex_u" , m_tex_u );
    data->get_value( "tex_v" , m_tex_v );
	data->get_value( "tex_w" , m_tex_w );
	
    m_squared_radius = sqr( m_radius );
    
    m_tex_u.normalize();
    m_tex_v.normalize();
    m_tex_w.normalize();
}    

//==============================================================================
//  Primitive_sphere::calculate_bounding_box()
//==============================================================================    
void Primitive_sphere::calculate_bounding_box() {
	m_bounding_box->set( get_position(), Vector3f(m_radius, m_radius, m_radius) );
}

//==============================================================================
//  Primitive_sphere::get_intersection()
//==============================================================================    
int Primitive_sphere::get_last_intersection( Ray& ray, float& dist, Intersection& intersection ) {
    Vector3f position = ray.get_origin();
    position += ray.get_direction() * dist;

    intersection.m_intersection = position;
    intersection.m_normal = (position - get_position()).get_normalized();
    intersection.m_primitive = this;
    intersection.m_material = m_material;
            
    Vector3f vp = intersection.m_normal;
	                
    float phi = acosf( -dot( vp, m_tex_w ) );
    intersection.m_v = phi * (1.0f / PI);
    intersection.m_u = (acosf( dot( m_tex_u, vp ) / sinf( phi ))) * (2.0f / PI);
    
	if( get_material()->used( Material::BUMP ) ) {
		intersection.m_tangent_t = cross ( cross( vp, m_tex_w ), vp );
        intersection.m_tangent_s = cross ( intersection.m_normal, intersection.m_tangent_t );
	}    
    
    return 1;
}

//==============================================================================
//  Primitive_sphere::get_intersection()
//==============================================================================    
Vector3f  Primitive_sphere::get_random_point_inside() {
     Vector3f sample_point;      
     float factor = 2.0f * m_radius / RAND_MAX;

     do {     
         sample_point.x = rand() * factor - m_radius;
         sample_point.y = rand() * factor - m_radius;
         sample_point.z = rand() * factor - m_radius;
     } while( sample_point.get_squared_length() > m_squared_radius );
     
     return get_position() + sample_point;
}

//==============================================================================
//  Primitive_plane::init()
//==============================================================================
void Primitive_plane::init(Data_node* data) {
    Primitive::init( data );

	m_tex_u = Vector3f( 1.0f, 0.0f, 0.0f );
	m_tex_v = Vector3f( 0.0f, 1.0f, 0.0f );

    data->get_value( "normal", m_normal );
    data->get_value( "distance", m_d );
    
    data->get_value( "tex_u" , m_tex_u );    
    data->get_value( "tex_v" , m_tex_v );
    
    m_tex_u.normalize();
    m_tex_v.normalize();
    
    create_batch();
}

//==============================================================================
//  Primitive_plane::get_intersection()
//==============================================================================    
int Primitive_plane::get_last_intersection( Ray& ray, float& dist, Intersection& intersection ) {
    Vector3f position = ray.get_origin();
    position += ray.get_direction() * dist;

    intersection.m_intersection = position;
                    
    intersection.m_normal = m_normal;
	intersection.m_tangent_s = m_tex_u;
	intersection.m_tangent_t = m_tex_v;
	
    intersection.m_primitive = this;
    intersection.set_material( m_material );
	
    intersection.m_u = dot( position, m_tex_u );
    intersection.m_v = dot( position, m_tex_v );
    return 1;
}

//==============================================================================
//  Primitive_plane::create_batch()
//==============================================================================    
void Primitive_plane::create_batch() {
	float* vertices = new float[ 4 * 5 ];
	unsigned short* indices = new unsigned short[ 6 ];
	
	indices[0] = 0;
	indices[1] = 1;
	indices[2] = 2;
	indices[3] = 2;
	indices[4] = 1;
	indices[5] = 3;
						
	m_batch = Factory_batch.create( "batch" );
	
	m_batch->set_vertices( vertices, 4, 5 * sizeof(float) );
	m_batch->set_indices( indices, 6 );
	
	m_batch->add_format( ATT_VERTEX, 3, 0 );
	m_batch->add_format( ATT_TEXCOORD, 2, 3 );
	
	Vector3f base_1 = cross( m_normal, Vector3f( m_normal.z, m_normal.x, -m_normal.y ));
	Vector3f base_2 = cross( m_normal, base_1 );
	
	log_debug << base_1 << base_2 << m_normal << std::endl;
//	base_1 = Vector3f( 100.0f, 0.0f, 0.0f );
//	base_2 = Vector3f( 0.0f, 100.0f, 0.0f );
			
	m_batch->get_vertex_by_index( 0 ) = (base_1+base_2) * 100.0f + m_normal*m_d;
	m_batch->get_texcoord_by_index( 0 ).x = dot( m_batch->get_vertex_by_index( 0 ), m_tex_u ) * m_material->m_tex_scale;
	m_batch->get_texcoord_by_index( 0 ).y = dot( m_batch->get_vertex_by_index( 0 ), m_tex_v ) * m_material->m_tex_scale;	
	m_batch->get_vertex_by_index( 1 ) = (base_1-base_2) * 100.0f + m_normal*m_d;
	m_batch->get_texcoord_by_index( 1 ).x = dot( m_batch->get_vertex_by_index( 1 ), m_tex_u ) * m_material->m_tex_scale;
	m_batch->get_texcoord_by_index( 1 ).y = dot( m_batch->get_vertex_by_index( 1 ), m_tex_v ) * m_material->m_tex_scale;	
	m_batch->get_vertex_by_index( 2 ) = (-base_1+base_2) * 100.0f + m_normal*m_d;
	m_batch->get_texcoord_by_index( 2 ).x = dot( m_batch->get_vertex_by_index( 2 ), m_tex_u ) * m_material->m_tex_scale;
	m_batch->get_texcoord_by_index( 2 ).y = dot( m_batch->get_vertex_by_index( 2 ), m_tex_v ) * m_material->m_tex_scale;	
	m_batch->get_vertex_by_index( 5 ) = (-base_1-base_2) * 100.0f + m_normal*m_d;
	m_batch->get_texcoord_by_index( 5 ).x = dot( m_batch->get_vertex_by_index( 5 ), m_tex_u ) * m_material->m_tex_scale;
	m_batch->get_texcoord_by_index( 5 ).y = dot( m_batch->get_vertex_by_index( 5 ), m_tex_v ) * m_material->m_tex_scale;	
	
	m_batch->add_normals();
}

//==============================================================================
//  Primitive_mesh::Primitive_mesh()
//==============================================================================
Primitive_mesh::Primitive_mesh() {
}

//==============================================================================
//  Primitive_mesh::Primitive_mesh()
//==============================================================================
Primitive_mesh::~Primitive_mesh() {
    for(int i=0, j=m_triangles.size(); i<j; ++i) {
        delete m_triangles[i];
    }
}

//==============================================================================
//  Primitive_mesh::init()
//==============================================================================          
void Primitive_mesh::init(Data_node* data) {
    Primitive::init( data );
    
	if( Data_value* value = data->get_value( "batch" ) ) {
		m_batch = Factory_batch.create( value );
		create_collision_set();
	}
}

//==============================================================================
//  Primitive_mesh::create_collision_set()
//==============================================================================          
bool Primitive_mesh::create_collision_set() {
	unsigned int num_indices = m_batch->get_index_count();
    bool find_normal   = m_batch->find_attribute( ATT_NORMAL );
    bool find_texcoord = m_batch->find_attribute( ATT_TEXCOORD );
    bool find_tangent  = m_batch->find_attribute( ATT_TANGENT_S );

    Vector3f v1, v2, v3;
    
    for(unsigned int i=0; i<num_indices; i+=3) {
         v1 = m_batch->get_vertex_by_index( i+0 );
         v2 = m_batch->get_vertex_by_index( i+1 );
         v3 = m_batch->get_vertex_by_index( i+2 );

         Triangle* triangle = new Triangle( v1, v2, v3 );

		 if( find_normal ) {
            triangle->m_normal_1  = m_batch->get_normal_by_index( i+0 );
            triangle->m_normal_21 = m_batch->get_normal_by_index( i+1 )-triangle->m_normal_1;
            triangle->m_normal_31 = m_batch->get_normal_by_index( i+2 )-triangle->m_normal_1;
         } else {
		   	Vector3f normal = cross( v2-v1, v3-v1).get_normalized();
			triangle->m_normal_1 = normal;
			triangle->m_normal_21 = triangle->m_normal_31 = Vector3f( 0.0f, 0.0f, 0.0f );
 		 }
 		  		 
		 if( find_texcoord ) {
		   	triangle->m_texcoord_1  = m_batch->get_texcoord_by_index( i+0 );
		   	triangle->m_texcoord_21 = m_batch->get_texcoord_by_index( i+1 )-triangle->m_texcoord_1;
		   	triangle->m_texcoord_31 = m_batch->get_texcoord_by_index( i+2 )-triangle->m_texcoord_1;
         } else {
		   	triangle->m_texcoord_1 = triangle->m_texcoord_21 = triangle->m_texcoord_31 = Vector2f( 0.0f, 0.0f );
 		 }
 		 
 		 if( find_tangent ) {
            triangle->m_tangent_s_1  = m_batch->get_tangent_s_by_index( i+0 );
            triangle->m_tangent_s_21 = m_batch->get_tangent_s_by_index( i+1 )-triangle->m_tangent_s_1;
            triangle->m_tangent_s_31 = m_batch->get_tangent_s_by_index( i+2 )-triangle->m_tangent_s_1;
		 }
		 
         m_triangles.push_back( triangle );
    }
    m_triangle_hit = m_triangles[0];
    
    m_batch->add_normals();
    return true;
}

//==============================================================================
//  Primitive_mesh::intersect( Ray& ray, float& dist )
//==============================================================================
int Primitive_mesh::intersect( Ray& ray, float& dist ) {
    if( !m_bounding_box->intersect( ray, dist ) ) return Ray::MISS;

    int hit, ret = Ray::MISS;

    for(int i=0, j=m_triangles.size(); i<j; ++i) {
        if( hit = m_triangles[i]->intersect( ray, dist ) ) {
            ret = hit;
            m_triangle_hit = m_triangles[i];
        }
    }
    return ret;
}

//==============================================================================
//  Primitive_mesh::shadow_intersect( Ray& ray, float& dist )
//==============================================================================
int Primitive_mesh::shadow_intersect( Ray& ray, float& dist ) {
    if( !m_bounding_box->intersect( ray, dist ) ) return Ray::MISS;
    
    for(int i=0, j=m_triangles.size(); i<j; ++i) {
        if( m_triangles[i]->intersect( ray, dist ) ) {
            return Ray::HIT;
        }
    }
    return Ray::MISS;
}

//==============================================================================
//  Primitive_mesh::get_intersection()
//==============================================================================
int Primitive_mesh::get_last_intersection( Ray& ray, float& dist, Intersection& intersection ) {
    intersection.m_primitive = this;
    intersection.set_material( m_material );

    Vector3f position = ray.get_origin();
    position += ray.get_direction() * dist;

    intersection.m_intersection = position;

    float m_u = m_triangle_hit->m_u * m_triangle_hit->ood;
    float m_v = m_triangle_hit->m_v * m_triangle_hit->ood;

    intersection.m_normal = m_triangle_hit->m_normal_1 + 
						  	m_v * m_triangle_hit->m_normal_21 + 
			 				m_u * m_triangle_hit->m_normal_31;
    intersection.m_normal.normalize();    
    
    Vector2f texcoord = m_triangle_hit->m_texcoord_1 + 
			 		    m_v * m_triangle_hit->m_texcoord_21 + 
						m_u * m_triangle_hit->m_texcoord_31;
						
	if( get_material()->used( Material::BUMP ) ) {
		intersection.m_tangent_s = m_triangle_hit->m_tangent_s_1 + 
                                   m_v * m_triangle_hit->m_tangent_s_21 + 
                                   m_u * m_triangle_hit->m_tangent_s_31;
        intersection.m_tangent_t = cross ( intersection.m_tangent_s, intersection.m_normal );
	}
		
	intersection.m_v = texcoord.y;
	intersection.m_u = texcoord.x;
}

//==============================================================================
//  Primitive_mesh::calculate_bounding_box()
//==============================================================================
void Primitive_mesh::calculate_bounding_box() {
	create_collision_set(); 
	 
    Vector3f m_min = Vector3f( FLOAT_MAX, FLOAT_MAX, FLOAT_MAX );
    Vector3f m_max = Vector3f( FLOAT_MIN, FLOAT_MIN, FLOAT_MIN );

    for(int i=0, j=m_triangles.size(); i<j; ++i) {
        Triangle* triangle = m_triangles[i];
        m_min = get_minima( triangle->m_v1, m_min );
        m_min = get_minima( triangle->m_v2, m_min );
        m_min = get_minima( triangle->m_v3, m_min );
        
        m_max = get_maxima( triangle->m_v1, m_max );
        m_max = get_maxima( triangle->m_v2, m_max );
        m_max = get_maxima( triangle->m_v3, m_max );
    }
    m_bounding_box->set_min_max( m_min, m_max );
}
