//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#ifndef _PRIMITIVE_
#define _PRIMITIVE_ 1

#include "../includes.hpp"
#include "../math/geometry.hpp"
#include "../misc/data.hpp"
#include "../misc/log.hpp"
#include "controller.hpp"

class Material;
class Primitive;
class Batch;

//==============================================================================
//  Intersection
//==============================================================================
class Intersection {
 public:
    Intersection() {
        m_u = m_v = 0.5f;
    }    
         
    Material* get_material() { return m_material; } 
    void set_material(Material* material) { m_material = material; }
               
public:
    Vector3f m_normal, m_intersection, m_tangent_s, m_tangent_t;
    Material* m_material;
    Primitive* m_primitive;
    float m_u, m_v; 
};

//==============================================================================
//  Primitive
//==============================================================================
class Primitive : public Controllable_object {
public:
    Primitive();
    ~Primitive();
     
    enum Type {
        PRIMITIVE,
        LIGHT
    };          
    enum Shape {
        UNKNOWN,
        POINT,
        SPHERE,
        PLANE,
        MESH
    };          
    enum Fall_off {
        NONE = 0,
        LINEAR = 1,
        QUADRATIC = 2
    };    
	 
    virtual void init(Data_node* data);

    virtual int intersect( Ray& ray, float& dist )=0;
    virtual int shadow_intersect( Ray& ray, float& dist ) { return intersect( ray, dist); }
    int get_intersection( Ray& ray, float& dist, Intersection& intersection ) {
		if( intersect( ray, dist ) ) {
			return get_last_intersection( ray, dist, intersection );
		}
		return Ray::MISS;
	}
    virtual int get_last_intersection( Ray& ray, float& dist, Intersection& intersection )=0;
    
    Material* get_material()                { return m_material; } 
    void set_material(Material* material)   { m_material = material; }
    
    Primitive::Type get_type()              { return m_type; }
    void set_type( Primitive::Type type)    { m_type = type; }
    
    virtual Primitive::Shape get_shape() {}
    
    AA_box* get_bounding_box()              { return m_bounding_box; }
    virtual void calculate_bounding_box()   {}
	        
//  Batch
    virtual void create_batch() {}
	Batch* get_batch()                      { return m_batch; }
	void set_batch( Batch* batch )          { m_batch = batch; }
	
//  Light functions     
    float get_intensity()                   { return m_intensity; }    
    void set_intensity( float intensity )   { m_intensity = intensity; }
    
    int get_max_samples()                   { return m_max_samples; }    
    void set_max_samples( int max_samples ) { m_max_samples = max_samples; }	
    
    int get_fall_off()                      { return m_fall_off; }
    void set_fall_off( Primitive::Fall_off fall_off )  { m_fall_off = fall_off; }
    
    bool get_cast_shadow()                  { return m_cast_shadow; }
    void set_cast_shadow(bool cast_shadow)  { m_cast_shadow = cast_shadow; }
    virtual Vector3f get_random_point_inside() { return get_position(); }
    
protected:
    AA_box* m_bounding_box;
    Batch* m_batch;
    Material* m_material;
    Primitive::Type m_type;
    bool m_cast_shadow;
    
    Fall_off m_fall_off; 
    float m_intensity;
    int m_max_samples;
private:
};

//==============================================================================
//  Primitive_point
//==============================================================================
class Primitive_point : public Primitive {
public:
	~Primitive_point() {}
	
    void init(Data_node* data);
    void calculate_bounding_box();
	        
    inline int intersect( Ray& ray, float& dist ) {
        return Ray::MISS;
    }
    
    int get_last_intersection( Ray& ray, float& dist, Intersection& intersection ) {};

    Primitive::Shape get_shape() { return Primitive::POINT; }    
};

//==============================================================================
//  Primitive_sphere
//==============================================================================
class Primitive_sphere : public Primitive {
public: 
	~Primitive_sphere() {}    
	
    void init(Data_node* data);
    void calculate_bounding_box();
         
    void set_radius( float radius ) { m_radius = radius; }
     
    inline int intersect( Ray& ray, float& dist ) {
		Vector3f v =  ray.get_origin() - get_position();
        float b =  -dot( v, ray.get_direction() );
        float det = (b * b) - dot(v, v) + (m_radius * m_radius);
    
        if (det > 0) {
            det = sqrtf( det );
            float i1 = b - det;
            float i2 = b + det;
        
            if (i2 > 0) {
                if (i1 < 0) {
                    if (i2 < dist) {
                        dist = i2;
                        return Ray::INPRIM;
                    }
                } else {
                    if (i1 < dist) {
                        dist = i1;
                        return Ray::HIT;
                    }
                }
            }
        }
        return Ray::MISS;
    }
    
    int get_last_intersection( Ray& ray, float& dist, Intersection& intersection );
    Vector3f get_random_point_inside();
        
    float&    get_radius()            { return m_radius; }
    float&    get_squared_radius()    { return m_squared_radius; }
    
    Primitive::Shape get_shape() { return Primitive::SPHERE; }    
protected:
    float m_radius, m_squared_radius;
    Vector3f m_tex_u, m_tex_v, m_tex_w;
};

//==============================================================================
//  Primitive_plane
//==============================================================================
class Primitive_plane : public Primitive {
public:
	~Primitive_plane() {}   
	   	          
    void init(Data_node* data);
      
    void set_normal( const  Vector3f& normal );
    void set_distance( float distance );
     
    inline int intersect( Ray& ray, float& dist ) {
        float det = dot( m_normal, ray.get_direction() );
        if( det < 0 ) {
            float distance = -(dot( m_normal, ray.get_origin() ) - m_d) / det;
            if ( distance > 0 && distance < dist) {
                dist = distance;
                return Ray::HIT;
            }
        }
        return Ray::MISS;  
    }
    int get_last_intersection( Ray& ray, float& dist, Intersection& intersection );
                   
    Primitive::Shape get_shape() { return Primitive::PLANE; } 
	void create_batch();   
protected:
	Vector3f m_normal;
	float m_d;
	Vector3f m_tex_u, m_tex_v;
};

//==============================================================================
//  Primitive_mesh
//==============================================================================
class Primitive_mesh : public Primitive {
public:
    Primitive_mesh();
    ~Primitive_mesh();
    
    void init(Data_node* data);
           
    int intersect( Ray& ray, float& dist );
    int get_last_intersection( Ray& ray, float& dist, Intersection& intersection );
    int shadow_intersect( Ray& ray, float& dist );
        
    Primitive::Shape get_shape() { return Primitive::MESH; }
    void calculate_bounding_box();
    bool create_collision_set();
protected:
    std::vector<Triangle* > m_triangles;
    Triangle* m_triangle_hit;
};

#endif /* _PRIMITIVE_ */
