//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "scene_portal.hpp"
#include "../../misc/log.hpp"
#include "../../misc/string_utils.hpp"
#include "../../misc/factory.hpp"
#include "../../math/vector.hpp"
#include "../../base.hpp"
#include "../camera.hpp"

#include "../primitive.hpp"
#include "../../material/material.hpp"
#include "../../material/shader.hpp"

#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <vector>

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_scene_portal = Factory_scene.reg< Scene_portal >( "scene_portal" );

//==============================================================================
//  Vertex_doom3
//==============================================================================
struct Vertex_doom3 {
    Vector3f vertex;
    Vector2f texcoord;
    Vector3f normal;
};

//==============================================================================
//  Scene_portal::render()
//==============================================================================
void Scene_portal::render() {
	Camera *camera = renderer->get_camera();

//  find start area..
    int start_area = get_area( camera->get_position() );
        
    std::stringstream caption;    
    caption << "Portal Engine, sector: " << (-1-start_area);
    caption << ", fps: " << base->get_fps();
    caption << " m cycles shaders, b, n toggles debug info)";
    base->set_caption( (caption.str()).c_str() );

	renderer->set_view( camera );
    renderer->set_renderport(0,0, base->get_window_width(), base->get_window_height());	
    renderer->clear_buffer( BUFFER_DEPTH | BUFFER_COLOR );
    
    if(start_area >= 0 || !m_use_portals) {
        // you're in the void!!
        renderer->render_scene( &m_primitives, &m_lights, NULL );
    } else {
        // rename start_area: -1 -> 0, -2 -> 1
        start_area = -1-start_area;        
        m_areas[start_area]->render_from_portal( camera, NULL, Vector2i(0, 0), Vector2i(base->get_window_width(), base->get_window_height()) );
    }       
}

//==============================================================================
//  Scene_portal::update()
//==============================================================================
bool Scene_portal::update() {
	if(base->is_key_released( KEY_B )) {
	    m_debug_light_aa_box = !m_debug_light_aa_box;
 	}
	if(base->is_key_released( KEY_N )) {
	    m_debug_light_quad = !m_debug_light_quad;
 	}
	if(base->is_key_released( KEY_M )) {
	    m_shader_index = (m_shader_index+1) % Factory_shader.get_objects().size();
 	}
	std::map<std::string, Shader*>::iterator it = Factory_shader.get_objects().begin();
	for(int i=0; i<m_shader_index; ++i) {
		it++;
	}
 	m_shader = (*it).second;
 	
    renderer->get_camera()->update();
    
    return true;
}

//==============================================================================
//  Portal_area::render_from_portal()
//==============================================================================
void Portal_area::render_from_portal(Camera* camera, Portal_portal* portal, Vector2i min, Vector2i max) {
    //  only render area once from portal each frame
    if(m_rendered[portal] == renderer->get_frame()) {
        return;
    }
	m_rendered[portal] = renderer->get_frame(); 

    //  set renderport, prevent visible 'rounding-error' caps
    renderer->set_renderport( min.x, min.y, max.x-min.x, max.y-min.y );
    
    int left = min.x, right = max.x, top = min.y, bottom = max.y;

	Shader* shader = m_scene->get_shader();

	// render to depthbuffer
	renderer->set_color( black );
	renderer->set_default();
	
    for(int i=0, s=m_primitives.size(); i<s; ++i) {
		if(camera->intersect( m_primitives[i]->get_bounding_box() ) ) {
			m_primitives[i]->get_batch()->render_solid( BUFFER_COLOR | BUFFER_DEPTH );
		}  
	}

	// render to colorbuffer, one light per pass

	renderer->set_blending( BLEND_ONE, BLEND_ONE );	
	renderer->set_shader( shader );
	
	shader->set_shader_global_constants();
	shader->set_shader_constants();

    AA_box bounding_box;		// bounding box of light
    Sphere bounding_sphere;		// bounding sphere of light
	
	for(int j=0, l=m_lights.size(); j<l; ++j) {
		Primitive* light = m_lights[j];
		float intensity = light->get_intensity();
				
		bounding_sphere.set_centre( light->get_position() );
		bounding_sphere.set_radius( intensity );	
		
		// intersect bounding sphere of light fustrum?
		if( camera->intersect( &bounding_sphere ) ) { 			
			int l,r,t,b;
			
			bounding_box.set( light->get_position(), Vector3f( intensity, intensity, intensity ) );		
			
			if( camera->project( &bounding_box, l, r, t, b ) ) {
				l = get_max(left, l); r = get_min(right, r); t = get_max(top, t); b = get_min(bottom, b);
		 	} else {
				l = left; r = right; t = top; b = bottom;			  	   
			}	
			
			if( (l < r) && (t < b) ) {
	   			renderer->set_scissor(l, t, r-l, b-t); 			
 			
			 	if(m_scene->get_debug_light_quad())   renderer->add_2d_debug_quad( Vector2f(l,t), Vector2f(r,b) );
				if(m_scene->get_debug_light_aa_box()) renderer->add_3d_debug_aa_box( &bounding_box );
		
				shader->set_shader_light_constants( light );

				for(int i=0, s=m_primitives.size(); i<s; ++i) {
				    Primitive* primitive = m_primitives[i];

				    // render primitive if primitive intersects boundingbox of light and fustrum
   					if(bounding_box.intersect(primitive->get_bounding_box()) && camera->intersect(primitive->get_bounding_box())) {
						shader->set_shader_material_constants( primitive->get_material() );
						primitive->get_batch()->render();
					}
				}
            }
        }
	}
	
	//  render portals
    for (int j=0, num_portals = m_portals.size(); j < num_portals; ++j) {
        if(m_portals[j] != portal) {
	        m_portals[j]->render_from_area( camera, m_index, min, max );
		}
    }
}

//==============================================================================
//  Portal_portal::render_from_area()
//==============================================================================
void Portal_portal::render_from_area(Camera* camera, int index, Vector2i min, Vector2i max) {
    // check if vertices are projected for visibility check    
    if(m_frame_rendered != renderer->get_frame()) {
        m_frame_rendered = renderer->get_frame();
        
        if( !(m_visible = check_visibility(camera)) ) {
            // portal is outside frustrum            
            return;
        }
        transform_points();
    }

    // check visibility
    if(!m_visible) {
        return;
    } else if(m_visible<0) {
    // intersection of portal and front plane of frustum
    // set min and max to renderport
        m_transformed_min = Vector2i(0,0);
        m_transformed_max = Vector2i(base->get_window_width(), base->get_window_height());    
    }
    
    // clip min and max
    if(min.x < m_transformed_min.x) min.x = m_transformed_min.x;
    if(max.x > m_transformed_max.x) max.x = m_transformed_max.x;
    
    if(min.y < m_transformed_min.y) min.y = m_transformed_min.y;
    if(max.y > m_transformed_max.y) max.y = m_transformed_max.y;
            
    // render area if visible
    if( (max.x > min.x) && (max.y > min.y) ) {
        if(index == m_area_pos) {
          m_scene->get_area( m_area_neg )->render_from_portal( camera, this, min, max );
        } else {
          m_scene->get_area( m_area_pos )->render_from_portal( camera, this, min, max );
        }    
    }    
}

//==============================================================================
//  Portal_portal::check_visibility()
//    0 = invisible (outside frustrum), 1 = visible, -1 = intersects frontplane
//==============================================================================
int Portal_portal::check_visibility(Camera* camera) {
    int num_points = m_points.size();
    int i;
    
    // check front plane.
    bool all_front = true;
    bool all_back  = true;
     
    for(i=0; i<num_points; ++i) {
        if(!camera->get_frustum_plane(FRUSTUM_FRONT).is_in_front(m_points[i])) {
            all_front = false;
        } else {
            all_back = false;
        }    
    }
    if(all_back) {
       return 0;
    }
    
    // check the other planes..    
    for(int j=FRUSTUM_FRONT+1; j<FRUSTUM_BACK; ++j) {
        bool all_back = true;
        for(i=0; i<num_points; ++i) {
            if(camera->get_frustum_plane(j).is_in_front(m_points[i])) {
                all_back = false;
                break;
            }  
        }
        if(all_back) {    
            return 0;
        }    
    }
    
    // return 1 if portal is in frustum and -1 if portal intersects frontplane
    return (all_front?1:-1);
}

//==============================================================================
//  Portal_portal::transform_points()
//==============================================================================
void Portal_portal::transform_points() {
    int num_points = m_points.size();
    
    m_transformed_min = Vector2i( 99999, 99999);
    m_transformed_max = Vector2i(-99999,-99999);
    
    for(int i=0; i<num_points; ++i) {
        if(renderer->get_camera()->project( m_points[i], m_transformed_points[i].x, m_transformed_points[i].y)) {
            // find maximum and minimum x, y and z values of transformed points to construct renderports.
            if(m_transformed_points[i].x > m_transformed_max.x) {
                m_transformed_max.x = m_transformed_points[i].x;
            } 
            if(m_transformed_points[i].x < m_transformed_min.x) {
                m_transformed_min.x = m_transformed_points[i].x;
            }
        
            if(m_transformed_points[i].y > m_transformed_max.y) {
                m_transformed_max.y = m_transformed_points[i].y;
            } 
            if(m_transformed_points[i].y < m_transformed_min.y) {
                m_transformed_min.y = m_transformed_points[i].y;
            }
        }    
    }
}

//==============================================================================
//  Scene_portal::get_area()
//==============================================================================
int Scene_portal::get_area(const Vector3f & position) {
    if(!m_nodes.size()) {
         return 0;
    }
    
    // walk through nodes...
    Doom3_node *node = &m_nodes[0];
    
    while(true) {	
        if( node->plane.is_in_front( position ) ) { // in front
            if( node->pos_child > 0 ) {
                node = &m_nodes[ node->pos_child ];
            } else {
                return node->pos_child;
              }        
         } else { // backside
            if( node->neg_child > 0 ) {
                node = &m_nodes[ node->neg_child ];                
            } else {
                return node->neg_child;
            }        
         }
    }
}

//==============================================================================
//  Scene_portal::push_light()
//==============================================================================
void Scene_portal::push_light(Primitive* light, Doom3_node *node) {
    if(!m_nodes.size()) {
    }
    
    // walk through nodes...
    if( node == NULL) node = &m_nodes[0];
    
    Vector3f position = light->get_position();
    float radius = light->get_intensity();
    float distance = node->plane.distance( position );
    
    if( distance > -radius ) { // front
        if( node->pos_child > 0 ) {
            push_light( light, &m_nodes[ node->pos_child ] );
        } else {
    	    // rename area: -1 -> 0, -2 -> 1
    	    int area = -1-node->pos_child;
    	    if( area >=0 ) {
			    m_areas[ area ]->add_light( light );
            }
        }
    }
	
	if( distance < radius ) { // backside
        if( node->neg_child > 0 ) {
            push_light( light, &m_nodes[ node->neg_child ] );
        } else {
	    // rename area: -1 -> 0, -2 -> 1
    	    int area = -1-node->neg_child;
    	    if( area >=0 ) {
			    m_areas[ area ]->add_light( light );
            }
        }        
   }
}

//==============================================================================
//  Scene_portal::get_area_index_by_name()
//==============================================================================
int Scene_portal::get_area_index_by_name(const std::string & name) {
    int num_areas = m_areas.size();
    
    for(int i=0; i<num_areas; ++i) {
        if(m_areas[i]->get_name() == name) {
               return i;
          }          
    }
    return -1;
}

//==============================================================================
//  Scene_portal::init()
//==============================================================================
void Scene_portal::init(Data_node* data) {
	Scene::init( data );
    
    std::string s, m;
        
    data->reset();
    while( Data_value* value = data->get_next_value() ) {
	    s = value->get_key();        
        if( s == "file_mtr" ) {
			value->read( m );
			load_mtr( m );
		}
	}
    
    data->get_value( "use_portals", m_use_portals);    
	if( data->get_value( "file_proc", s ) ) {
		load_proc( s );
    }
	if( data->get_value( "file_map", s ) ) {
		load_map( s );
    }	
}

//============================================================================//
//                                                                            //
//    Read from file functions (format: .proc, doom3)                         //
//                                                                            //
//============================================================================//
inline std::string proc_get_next_value(std::ifstream & file) {
    std::string s;
    while( file >> s) {
        if( s == "/*" ) {
            while( s != "*/") {
                file >> s;
            }    
        } else if( s == "{" || s == "}") {
        } else if( s == "(" || s == ")") {
        } else {
            return s;
        }             
    }
    return false;
}  

inline std::string proc_get_next_string(std::ifstream & file) {
    std::string s = proc_get_next_value(file);
    
    return s.substr(1, s.size()-2);
}

#define proc_get_next_float(x) atof( proc_get_next_value(x).c_str() )
#define proc_get_next_int(x) atoi( proc_get_next_value(x).c_str() ) 

//==============================================================================
//  Scene_portal::load_proc()
//==============================================================================
void Scene_portal::load_proc(const std::string & name) {
    log_function << "Load " << name << " in Scene_portal::load_proc()" << std::endl;
    
    std::ifstream file(name.c_str());
    std::string s;
      
    if(!file) {
        log_error << "unable to open " << name << std::endl;
        return;
    }
    
    while( file >> s) {
        if(s == "model") {
// areas
            std::string name = proc_get_next_string(file);
            Portal_area* portal_area = new Portal_area(this, name, m_areas.size());
            portal_area->read_from_file(file);
            m_areas.push_back(portal_area);
        } else if (s == "interAreaPortals") {
// portals
            int num_areas   = proc_get_next_int(file);
            int num_portals = proc_get_next_int(file);

            log_debug << "load " << num_portals << " portals" << std::endl;

            m_portals.resize(num_portals);

            for(int i=0; i<num_portals; ++i) {
                Portal_portal* portal_portal = new Portal_portal(this);
                portal_portal->read_from_file(file);
                m_portals[i] = portal_portal;
            }    
        } else if (s == "nodes") {
// nodes
            int num_nodes = proc_get_next_int(file);
      
            log_debug << "load " << num_nodes << " nodes" << std::endl;
            
            m_nodes.resize(num_nodes);
            
            for(int i=0; i<num_nodes; ++i) {
                Doom3_node node;
                node.plane.m_normal.x = proc_get_next_float(file);
                node.plane.m_normal.y = proc_get_next_float(file);
                node.plane.m_normal.z = proc_get_next_float(file);
                node.plane.m_d = proc_get_next_float(file);
                
                node.pos_child = proc_get_next_int(file);
                node.neg_child = proc_get_next_int(file);

                if(node.pos_child<0) {
                    std::stringstream name;
                    name << "_area" << (-1-node.pos_child);
                    node.pos_child = -1-get_area_index_by_name(name.str());
                   }

                if(node.neg_child<0) {
                    std::stringstream name;
                    name << "_area" << (-1-node.neg_child);
                    node.neg_child = -1-get_area_index_by_name(name.str());
                }
                   
                m_nodes[i] = node;
            }
        }    
    }
    file.close();
    
    log_success << name << " loaded succesfully " << std::endl;
}
    
//==============================================================================
//  Portal_portal::read_from_file()
//==============================================================================
void Portal_portal::read_from_file(std::ifstream &file) {
    int num_points = proc_get_next_int( file );
    int pos_area   = proc_get_next_int( file );
    int neg_area   = proc_get_next_int( file );    
    
    m_points.resize(num_points);
    m_transformed_points.resize(num_points);
    
    for(int i=0; i<num_points; ++i) {
        Vector3f tmp;
        tmp.x = proc_get_next_float( file );
        tmp.y = proc_get_next_float( file );
        tmp.z = proc_get_next_float( file );                
        
        m_points[i] = tmp;
    }
    
    std::stringstream name_pos;
    name_pos << "_area" << pos_area;
    m_area_pos = m_scene->get_area_index_by_name(name_pos.str());
    std::stringstream name_neg;    
    name_neg << "_area" << neg_area;
    m_area_neg = m_scene->get_area_index_by_name(name_neg.str());    
    
    if(m_area_pos>=0) {
        m_scene->get_area(m_area_pos)->add_portal(this);
    }    
    if(m_area_neg>=0) {
        m_scene->get_area(m_area_neg)->add_portal(this);
    }    
}

//==============================================================================
//  Portal_area::read_from_file()
//==============================================================================
void Portal_area::read_from_file(std::ifstream &file) {
    log_function << "Load area (model) " << m_name << " in Portal_area::read_from_file()" << std::endl;
       // read num surfaces;               
    int num_surfaces = proc_get_next_int(file);
    log_debug << num_surfaces << " surface(s)" << std::endl;

    for(int i=0; i<num_surfaces; i++) {        
		// create new primitive ..
		Primitive* primitive = Factory_primitive.create( "mesh" );	
        // create new batch...
        Batch* batch = Factory_batch.create( "batch" );
        batch->add_format( ATT_VERTEX,   3, 0, 0);
        batch->add_format( ATT_TEXCOORD, 2, 3, 0);
        batch->add_format( ATT_NORMAL,   3, 5, 0);                     
        batch->set_primitive_type(PRIM_TRIANGLES);
        std::vector<Vertex_doom3> *vertices = new std::vector<Vertex_doom3>();                
        std::vector<unsigned short> *indices = new std::vector<unsigned short>();
                
        // read surface...
        std::string mat_name = proc_get_next_string(file);
                
        int num_verts = proc_get_next_int(file);
        int num_ind = proc_get_next_int(file);             
            
        vertices->resize( num_verts );
        indices->resize( num_ind );
            
        for(int j=0; j<num_verts; j++) {
            Vertex_doom3 tmp_vertex;
            tmp_vertex.vertex.x = atof( proc_get_next_value(file).c_str() );
            tmp_vertex.vertex.y = atof( proc_get_next_value(file).c_str() );
            tmp_vertex.vertex.z = atof( proc_get_next_value(file).c_str() );

            tmp_vertex.texcoord.x = atof( proc_get_next_value(file).c_str() );
            tmp_vertex.texcoord.y = atof( proc_get_next_value(file).c_str() );

            tmp_vertex.normal.x = atof( proc_get_next_value(file).c_str() );
            tmp_vertex.normal.y = atof( proc_get_next_value(file).c_str() );
            tmp_vertex.normal.z = atof( proc_get_next_value(file).c_str() );

            vertices->at(j) = (tmp_vertex);
        }
        for(int j=0; j<num_ind; j++) {
            indices->at(j) = ( atoi( proc_get_next_value(file).c_str() ) );
        }
        for(int j=0; j<num_ind; j+=3) {
			if( dot( cross( vertices->at(indices->at(j+1)).vertex-vertices->at(indices->at(j)).vertex, 
				            vertices->at(indices->at(j+2)).vertex-vertices->at(indices->at(j)).vertex ), vertices->at(indices->at(j)).normal ) < 0 ) {
				int tmp = indices->at(j+1);
                indices->at(j+1) = indices->at(j+2);
                indices->at(j+2) = tmp;
			}
        }          
      
        // save batch..
        batch->set_vertices( &vertices->at(0), vertices->size(), sizeof(Vertex_doom3));
        batch->set_indices( &indices->at(0), indices->size() );                          

		batch->add_normals();
		batch->add_tangent_space();

		// init primitive
		Material* material = Factory_material.use( mat_name );
		
		primitive->set_material( material );
		primitive->set_batch( batch );
		
		if( material ) {		
			/* load images needed */
			material->get_texture( Material::DIFFUSE )->set_image( get_image_from_file( material->get_texture( Material::DIFFUSE )->get_name() ) );
			material->get_texture( Material::BUMP )->set_image( get_image_from_file( material->get_texture( Material::BUMP )->get_name() ) );			
			material->get_texture( Material::SPECULAR )->set_image( get_image_from_file( material->get_texture( Material::SPECULAR )->get_name() ) );
			
		    m_primitives.push_back( primitive );		
		    m_scene->m_primitives.push_back( primitive );
		    m_scene->m_batches.push_back( batch );
		} else {
//		  	log_debug << "Material " << mat_name << " not found " << std::endl;   
  		}
    }
    log_success << m_name << " loaded succesfully " << std::endl;    
}

//============================================================================//
//                                                                            //
//    Read from file functions (format: .map, doom3)                          //
//	  - only light entities at the moment.                                    //
//                                                                            //
//============================================================================//

//==============================================================================
//  Scene_portal::load_map()
//==============================================================================
void Scene_portal::load_map(const std::string & name) {	 
    log_function << "Load " << name << " in Scene_portal::load_map()" << std::endl;
    
    std::ifstream file(name.c_str());
    std::string s;
      
    if(!file) {
        log_error << "unable to open " << name << std::endl;
        return;
    }
    
    while( file >> s) {
        if(s == "\"light\"") {
		    // new light...			 
			Primitive* light = Factory_primitive.create( "point" );
            light->set_type( Primitive::LIGHT );
//			light->set_fall_off( Primitive::LINEAR );
  	 		std::string mat("1 1 1");
  	 		Material* material = NULL;
			float intensity = 0;
			    
  	 		while( s != "}" && file >> s ) {
			    if( s == "\"origin\"" ) {
				    Vector3f origin;
				    unsigned char c1, c2;
				 	file >> c1 >> origin >> c2;
				 	light->get_position() = origin;
			 	} else if( s == "\"_color\"" ) {
				    Vector3f color;
				    unsigned char c1, c2;
				 	file >> c1 >> color >> c2;	
					std::ostringstream oss;
					oss << color;
					material = Factory_material.use( oss.str() );
					if(!material) {
					    material = Factory_material.create( "material", oss.str() );	
					    material->set_color( Material::DIFFUSE, color );
					}
					light->set_material( material );
	   		    } else if( s == "\"light_radius\"") {
		   		    Vector3f radius;
				    unsigned char c1, c2;
				 	file >> c1 >> radius >> c2;				 	
				 	intensity = (radius.x+radius.y+radius.z)/3.0f;
				 	if( intensity < 0.1f ) {
						log_fatal << intensity << std::endl;
						intensity = 1.0f;
					}
				 	light->set_intensity( intensity );
	   		    }
            }  	 		 
            if(!material || !intensity) {
			//	delete light;		 
			} else {
			    light->first_update();
                push_light(light);
			}	
	    } else if( s == "\"info_player_start\"" ) {
		  	while( s != "}" && file >> s ) {
			    if( s == "\"origin\"" ) {
				    Vector3f origin;
				    unsigned char c1, c2;
				 	file >> c1 >> origin >> c2;
				 	if(Camera* camera = renderer->get_camera() ) {
						camera->get_controller()->look_at( origin, origin+Vector3f(0,1,0), Vector3f(0,0,1) );
						camera->first_update();
					}
			 	}   
			}	
		}
    }
    file.close();

	for(int i=0; i<2; ++i) {
		Primitive* light = Factory_primitive.create( "point" );
		light->set_type( Primitive::LIGHT );
		light->get_position() = Vector3f(5000.0f*(float)(2*i-1), 0.0f, 5000.0f*(float)(2*i-1));
		Material *material = Factory_material.create( "material" );
	    material->set_color( Material::DIFFUSE, Vector3f( (float)(i+1)*0.5f, 0.9f, 0.8f ) );
		light->set_material( material );
	 	light->set_intensity( 15000.0f );
	    light->first_update();
	    m_lights.push_back(light);
	}
    log_success << name << " loaded succesfully " << std::endl;
}

//============================================================================//
//                                                                            //
//    Read from file functions (format: .mtr, doom3)                          //
//	  - only diffusemap, bumpmap and specularmap at the moment.               //
//                                                                            //
//============================================================================//

//==============================================================================
//  Scene_portal::load_mtr()
//==============================================================================

std::string get_texture_name( std::ifstream &file, const std::string alt, const std::string alt2 ) {
    std::string s;
    
    file >> s;
  	if( s == "map" || s == "Map" ) {
		file >> s;
	}
  	if( s == "addnormals(" || s == "addnormals") {
		file >> s;
	}
	
	find_and_replace( s, std::string(" "), std::string("") );
	find_and_replace( s, std::string(","), std::string("") );
	find_and_replace( s, std::string("\\"), std::string("/") );
	find_and_replace( s, std::string("addnormals("), std::string("") );
	find_and_replace( s, std::string("("), std::string("") );
	
	s = trim( s );
	
	int pos = s.find( ".tga" );
	if( pos<0 ) {
		s.append(".tga");
	}

	std::ifstream chk( s.c_str() );
	if( !chk )  {
		s = alt; 
		std::ifstream chk2( s.c_str() );
		if(!chk2) {
			s = alt2;	  
		} else {
		  	chk2.close();   
	    }
	} else chk.close();
		
	return s;
}

void Scene_portal::load_mtr(const std::string & name) {	 
    log_function << "Load " << name << " in Scene_portal::load_mtr()" << std::endl;
    
    std::ifstream file(name.c_str());
    std::string s, bm, gm, dm, mat_name;
      
    if(!file) {
        log_error << "unable to open " << name << std::endl;
        return;
    }
    
    int rec = 0;    
    Texture *bumpmap, *diffusemap, *specularmap;
        
	specularmap = get_texture_from_file("../data/textures/flat_reflection.tga");
	diffusemap  = get_texture_from_file("default");
	bumpmap     = get_texture_from_file("../data/textures/flat_local.tga");
    
    while( file >> s) {
		
		   
        if(s[0] == '/' && s[1] == '/') {
			getline(file,s,'\n');			 
		} else if( s == "{" ) {
  		    rec++;
  		} else if( rec == 0 ) {
			mat_name = s; 
			
			dm = mat_name;
			dm.append(".tga");
			gm = mat_name;
			gm.append("_s.tga");
			bm = mat_name;
			bm.append("_local.tga");
									
	    } else if( s == "}" ) {
  		    rec--;
  	  	 		    
  		    if( rec == 0 ) {
		    	Material* material = Factory_material.create( "material", mat_name );
		    	
   				bumpmap = Factory_texture.create( "texture_2d", bm );
				bumpmap->set_name( bm );
				diffusemap = Factory_texture.create( "texture_2d", dm );
				diffusemap->set_name( dm );
				specularmap = Factory_texture.create( "texture_2d", gm );
				specularmap->set_name( gm );
		    	
				material->set_texture( Material::BUMP, bumpmap );
				material->set_texture( Material::DIFFUSE, diffusemap );
				material->set_texture( Material::SPECULAR, specularmap );
			}
			
  		} else if( s == "bumpmap" ) {
		  	bm = get_texture_name( file, bm, std::string("../data/textures/flat_local.tga") );
	    } else if( s == "diffusemap" || s == "diffuseMap") {
		  	dm = get_texture_name( file, dm, std::string("default") );
		} else if( s == "specularmap") {
		  	gm = get_texture_name( file, gm, std::string("../data/textures/flat_reflection.tga") );
  		}  		
    }
    file.close();

    log_success << name << " loaded succesfully " << std::endl;
}
